#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pyomo.environ as pyo


__author__ = "Alexander V. Dudchenko (SLAC)"


class ModelManager:
    def __init__(self, ps_instance):
        """This class manages the model state
        It uses the ps_instance info so it can manage
        model build, init and solve logic, as well as get any other config paramters
        """
        self.ps_conf = ps_instance.config
        self.ps = ps_instance

        self.is_initialized = False
        self.is_solved = False
        self.is_prior_parameter_solved = False
        # internal state option to be used by differntial sweep tool to
        # prevent model reintialization and state reset
        self._is_rebuild_and_init_enabled = True
        # this is isused for loggin states if enabled
        self.solved_states = {"state": [], "local_value_k": []}
        self.initialized_states = {"state": [], "local_value_k": []}
        self.current_k = None

    def build_and_init(self, sweep_params=None, local_value_k=None):
        """build and init model, if required by user, update paramaters before init"""
        self.model = self.ps_conf.build_model(**self.ps_conf.build_model_kwargs)
        # intilized model if init function is passed in
        if self.ps_conf.initialize_function is not None:
            # update paramters before init if enabled by user
            if (
                self.ps_conf.update_sweep_params_before_init
                and sweep_params is not None
                and local_value_k is not None
            ):
                self.update_model_params(sweep_params, local_value_k)
            # init

            self.init_model()

        # raise error if user sets to init before sweep, but does not provide
        # initilize function
        elif self.ps_conf.update_sweep_params_before_init:
            raise ValueError(
                "Initialization function was not specified. The model will not be reinitialized with specified paramters."
            )
        self.update_initialized_state(True)
        self.update_solved_state(False)

    def update_initialized_state(self, state):
        """used to update and if requested log intilized state"""
        if self.ps_conf.log_model_states:
            self.initialized_states["state"].append(state)
            self.initialized_states["local_value_k"].append(self.current_k)

        self.is_initialized = state

    def update_solved_state(self, state):
        """used to update and if requested log solved state"""
        if self.ps_conf.log_model_states:
            self.solved_states["state"].append(state)
            self.solved_states["local_value_k"].append(self.current_k)
        self.is_solved = state

    def init_model(self):
        """attempt to init model"""
        try:
            self.ps_conf.initialize_function(
                self.model, **self.ps_conf.initialize_kwargs
            )
            self.update_initialized_state(True)
        except TypeError:
            # this happens if the optimize_kwargs are misspecified,
            # which is an error we want to raise
            self.update_solved_state(False)
            self.update_initialized_state(False)
            raise
        except:
            self.update_solved_state(False)
            self.update_initialized_state(False)

    def update_model_params(self, sweep_params, local_value_k):
        """Update local params, and store current local k state"""
        # print(self.model, sweep_params, local_value_k)
        self.ps._update_model_values(self.model, sweep_params, local_value_k)
        self.current_k = local_value_k

    def solve_model(self):
        """Attempt to solve the unit model"""
        # update to determine if we are solving from initilized or pre-solved state
        self.is_prior_parameter_solved = self.is_solved
        try:
            results = self.ps_conf.optimize_function(
                self.model, **self.ps_conf.optimize_kwargs
            )
            pyo.assert_optimal_termination(results)
            self.update_solved_state(True)
            self.update_initialized_state(True)
            return results
        except TypeError:
            # this happens if the optimize_kwargs are misspecified,
            # which is an error we want to raise
            self.update_solved_state(False)
            self.update_initialized_state(False)
            raise
        except:
            self.update_solved_state(False)
            self.update_initialized_state(False)
        return None



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains common methods for determining the electricity intensity
and demand for zero-order unit models.
"""

import idaes.logger as idaeslog

from pyomo.environ import Param, Var, units as pyunits

# Some more inforation about this module
__author__ = "Andrew Lee"

# Set up logger
_log = idaeslog.getLogger(__name__)


def _common(self):
    # Add electricity consumption to model
    self.electricity = Var(
        self.flowsheet().time,
        units=pyunits.kW,
        bounds=(0, None),
        doc="Electricity consumption of unit",
    )

    self._perf_var_dict["Electricity Demand"] = self.electricity


def constant_intensity(self):
    """
    Helper method for implementing electricity demand assuming constant
    intensity based on the inlet volumetric flow rate.

    E[t] = Q[t] * intensity

    Two variables are added to the model:
        * electricity (indexed by time)
        * energy_electric_flow_vol_inlet (unindexed)

    One constraint is added to the model:
        * electricity_consumption (indexed by time)
    """
    _common(self)

    self.energy_electric_flow_vol_inlet = Var(
        units=pyunits.kWh / pyunits.m**3,
        doc="Electricity intensity with respect to inlet flowrate of unit",
    )

    @self.Constraint(
        self.flowsheet().time,
        doc="Constraint for electricity consumption based on " "feed flowrate.",
    )
    def electricity_consumption(b, t):
        return b.electricity[t] == (
            b.energy_electric_flow_vol_inlet
            * pyunits.convert(
                b.get_inlet_flow(t), to_units=pyunits.m**3 / pyunits.hour
            )
        )

    self._fixed_perf_vars.append(self.energy_electric_flow_vol_inlet)
    self._perf_var_dict["Electricity Intensity"] = self.energy_electric_flow_vol_inlet


def pump_electricity(self, flow_rate):
    """
    Helper method for calculating electricity demand based on a
    pump flow equation.

    E[t] = (0.746 * Q[t] * H / (3960 * eta_pump * eta_motor))

    Here Q is the volumetric flowrate to be used to calculate electricity demand,
    H is the lift height for the pump and eta_pump and eta_motor are the
    efficiencies of the pump and motor respectively.

    Args:
        flow_rate - term to use for Q in the equation above.
    """
    _common(self)

    self.lift_height = Param(
        initialize=100, units=pyunits.feet, mutable=True, doc="Lift height for pump"
    )
    self.eta_pump = Param(
        initialize=0.9,
        units=pyunits.dimensionless,
        mutable=True,
        doc="Efficiency of pump",
    )
    self.eta_motor = Param(
        initialize=0.9,
        units=pyunits.dimensionless,
        mutable=True,
        doc="Efficiency of motor",
    )

    @self.Constraint(
        self.flowsheet().time,
        doc="Constraint for electricity consumption based on " "pump flowrate.",
    )
    def electricity_consumption(b, t):
        A = 3960 * pyunits.gallon * pyunits.foot / pyunits.minute / pyunits.horsepower
        return b.electricity[t] == pyunits.convert(
            flow_rate[t] * self.lift_height / (A * self.eta_pump * self.eta_motor),
            to_units=pyunits.kW,
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
from watertap.tools.parameter_sweep.sampling_types import (
    SamplingType,
    LinearSample,
    GeomSample,
    ReverseGeomSample,
    UniformSample,
    NormalSample,
    LatinHypercubeSample,
)
from watertap.tools.parameter_sweep.parameter_sweep_functions import (
    parameter_sweep,
    recursive_parameter_sweep,
    differential_parameter_sweep,
)
from watertap.tools.parameter_sweep.parameter_sweep_reader import (
    get_sweep_params_from_yaml,
    set_defaults_from_yaml,
    ParameterSweepReader,
)
from watertap.tools.parameter_sweep.parameter_sweep import (
    ParameterSweep,
    RecursiveParameterSweep,
)

from watertap.tools.parameter_sweep.parameter_sweep_differential import (
    DifferentialParameterSweep,
)

# TODO: should this be removed?
import numpy as _np

_np.set_printoptions(linewidth=200)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest
from pyomo.environ import (
    ConcreteModel,
    Block,
    Var,
    Constraint,
    Expression,
    value,
    Objective,
    assert_optimal_termination,
)
from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.models.unit_models import Feed, Product, Mixer, Separator
from idaes.models.unit_models.heat_exchanger import HeatExchanger
from idaes.models.unit_models.translator import Translator
from pyomo.util.check_units import assert_units_consistent
from idaes.core.util.model_statistics import degrees_of_freedom, number_total_objectives

from watertap.examples.flowsheets.mvc.mvc_single_stage import (
    build,
    set_operating_conditions,
    add_Q_ext,
    initialize_system,
    scale_costs,
    fix_outlet_pressures,
    solve,
    set_up_optimization,
)

import watertap.property_models.seawater_prop_pack as props_seawater
import watertap.property_models.water_prop_pack as props_water
from watertap.unit_models.pressure_changer import Pump
from watertap.unit_models.mvc.components import Evaporator, Compressor, Condenser


solver = get_solver()


class TestMVC:
    @pytest.fixture(scope="class")
    def mvc_single_stage(self):
        m = build()
        add_Q_ext(m, time_point=m.fs.config.time)
        return m

    @pytest.mark.unit
    def test_build_model(self, mvc_single_stage):
        m = mvc_single_stage

        # test model set up
        assert isinstance(m, ConcreteModel)
        assert isinstance(m.fs, FlowsheetBlock)
        assert isinstance(m.fs.properties_feed, props_seawater.SeawaterParameterBlock)
        assert isinstance(m.fs.properties_vapor, props_water.WaterParameterBlock)
        assert isinstance(m.fs.costing, Block)

        # test unit models
        assert isinstance(m.fs.feed, Feed)
        assert isinstance(m.fs.distillate, Product)
        assert isinstance(m.fs.brine, Product)
        assert isinstance(m.fs.pump_feed, Pump)
        assert isinstance(m.fs.pump_distillate, Pump)
        assert isinstance(m.fs.pump_brine, Pump)
        assert isinstance(m.fs.separator_feed, Separator)
        assert isinstance(m.fs.hx_distillate, HeatExchanger)
        assert isinstance(m.fs.hx_brine, HeatExchanger)
        assert isinstance(m.fs.mixer_feed, Mixer)
        assert isinstance(m.fs.evaporator, Evaporator)
        assert isinstance(m.fs.compressor, Compressor)
        assert isinstance(m.fs.condenser, Condenser)
        assert isinstance(m.fs.tb_distillate, Translator)

        # unit model options
        assert isinstance(m.fs.hx_distillate.cold.deltaP, Var)
        assert isinstance(m.fs.hx_distillate.hot.deltaP, Var)
        assert isinstance(m.fs.hx_brine.cold.deltaP, Var)
        assert isinstance(m.fs.hx_brine.hot.deltaP, Var)

        # additional constraints, variables, and expressions
        assert isinstance(m.fs.recovery, Var)
        assert isinstance(m.fs.recovery_equation, Constraint)
        assert isinstance(m.fs.split_ratio_recovery_equality, Constraint)
        assert isinstance(m.fs.Q_ext, Var)
        assert isinstance(m.fs.costing.annual_water_production, Expression)
        assert isinstance(m.fs.costing.specific_energy_consumption, Expression)

        # costing blocks
        for blk_str in ("evaporator", "compressor", "hx_distillate", "hx_brine"):
            blk = getattr(m.fs, blk_str)
            c_blk = getattr(blk, "costing")
            assert isinstance(c_blk, Block)
            assert isinstance(getattr(c_blk, "capital_cost"), Var)

        var_str_list = [
            "total_capital_cost",
            "total_operating_cost",
        ]
        for var_str in var_str_list:
            var = getattr(m.fs.costing, var_str)
            assert isinstance(var, Var)

        # arcs
        arc_dict = {
            m.fs.s01: (m.fs.feed.outlet, m.fs.pump_feed.inlet),
            m.fs.s02: (m.fs.pump_feed.outlet, m.fs.separator_feed.inlet),
            m.fs.s03: (
                m.fs.separator_feed.hx_distillate_cold,
                m.fs.hx_distillate.cold_inlet,
            ),
            m.fs.s04: (m.fs.separator_feed.hx_brine_cold, m.fs.hx_brine.cold_inlet),
            m.fs.s05: (
                m.fs.hx_distillate.cold_outlet,
                m.fs.mixer_feed.hx_distillate_cold,
            ),
            m.fs.s06: (m.fs.hx_brine.cold_outlet, m.fs.mixer_feed.hx_brine_cold),
            m.fs.s07: (m.fs.mixer_feed.outlet, m.fs.evaporator.inlet_feed),
            m.fs.s08: (m.fs.evaporator.outlet_vapor, m.fs.compressor.inlet),
            m.fs.s09: (m.fs.compressor.outlet, m.fs.condenser.inlet),
            m.fs.s10: (m.fs.evaporator.outlet_brine, m.fs.pump_brine.inlet),
            m.fs.s11: (m.fs.pump_brine.outlet, m.fs.hx_brine.hot_inlet),
            m.fs.s12: (m.fs.hx_brine.hot_outlet, m.fs.brine.inlet),
            m.fs.s13: (m.fs.condenser.outlet, m.fs.tb_distillate.inlet),
            m.fs.s14: (m.fs.tb_distillate.outlet, m.fs.pump_distillate.inlet),
            m.fs.s15: (m.fs.pump_distillate.outlet, m.fs.hx_distillate.hot_inlet),
            m.fs.s16: (m.fs.hx_distillate.hot_outlet, m.fs.distillate.inlet),
        }
        for arc, port_tpl in arc_dict.items():
            assert arc.source is port_tpl[0]
            assert arc.destination is port_tpl[1]

        # units
        assert_units_consistent(m.fs)

    @pytest.mark.component
    def test_set_operating_conditions(self, mvc_single_stage):
        m = mvc_single_stage
        set_operating_conditions(m)

        # check fixed variables
        # feed
        assert m.fs.feed.properties[0].mass_frac_phase_comp["Liq", "TDS"].is_fixed()
        assert value(m.fs.feed.properties[0].mass_frac_phase_comp["Liq", "TDS"]) == 0.1
        assert m.fs.feed.properties[0].flow_mass_phase_comp["Liq", "H2O"].is_fixed()
        assert value(m.fs.feed.properties[0].flow_mass_phase_comp["Liq", "H2O"]) == 40
        assert m.fs.feed.temperature[0].is_fixed()
        assert value(m.fs.feed.temperature[0]) == 298.15
        assert m.fs.feed.pressure[0].is_fixed()
        assert value(m.fs.feed.pressure[0]) == 101325

        # recovery
        assert m.fs.recovery[0].is_fixed()
        assert value(m.fs.recovery[0]) == 0.5

        # pumps
        assert m.fs.pump_feed.efficiency_pump[0].is_fixed()
        assert value(m.fs.pump_feed.efficiency_pump[0]) == 0.8
        assert m.fs.pump_feed.control_volume.deltaP[0].is_fixed()
        assert value(m.fs.pump_feed.control_volume.deltaP[0]) == 7e3

        assert m.fs.pump_distillate.efficiency_pump[0].is_fixed()
        assert value(m.fs.pump_distillate.efficiency_pump[0]) == 0.8
        assert m.fs.pump_distillate.control_volume.deltaP[0].is_fixed()
        assert value(m.fs.pump_distillate.control_volume.deltaP[0]) == 4e4

        assert m.fs.pump_brine.efficiency_pump[0].is_fixed()
        assert value(m.fs.pump_brine.efficiency_pump[0]) == 0.8
        assert m.fs.pump_brine.control_volume.deltaP[0].is_fixed()
        assert value(m.fs.pump_brine.control_volume.deltaP[0]) == 4e4

        # heat exchangers
        assert m.fs.hx_distillate.overall_heat_transfer_coefficient[0].is_fixed()
        assert value(m.fs.hx_distillate.overall_heat_transfer_coefficient[0]) == 2e3
        assert m.fs.hx_distillate.area.is_fixed()
        assert value(m.fs.hx_distillate.area) == 125
        assert m.fs.hx_distillate.cold.deltaP[0].is_fixed()
        assert value(m.fs.hx_distillate.cold.deltaP[0]) == 7e3
        assert m.fs.hx_distillate.hot.deltaP[0].is_fixed()
        assert value(m.fs.hx_distillate.hot.deltaP[0]) == 7e3

        assert m.fs.hx_brine.overall_heat_transfer_coefficient[0].is_fixed()
        assert value(m.fs.hx_brine.overall_heat_transfer_coefficient[0]) == 2e3
        assert m.fs.hx_brine.area.is_fixed()
        assert value(m.fs.hx_brine.area) == 115
        assert m.fs.hx_brine.cold.deltaP[0].is_fixed()
        assert value(m.fs.hx_brine.cold.deltaP[0]) == 7e3
        assert m.fs.hx_brine.hot.deltaP[0].is_fixed()
        assert value(m.fs.hx_brine.hot.deltaP[0]) == 7e3

        # evaporator
        assert m.fs.evaporator.outlet_brine.temperature[0].is_fixed()
        assert value(m.fs.evaporator.outlet_brine.temperature[0]) == 343.15
        assert m.fs.evaporator.U.is_fixed()
        assert value(m.fs.evaporator.U) == 3e3

        # compressor
        assert m.fs.compressor.pressure_ratio.is_fixed()
        assert value(m.fs.compressor.pressure_ratio) == 1.6
        assert m.fs.compressor.efficiency.is_fixed()
        assert value(m.fs.compressor.efficiency) == 0.8

        # 0 TDS in distillate
        assert (
            m.fs.tb_distillate.properties_out[0]
            .flow_mass_phase_comp["Liq", "TDS"]
            .is_fixed()
        )
        assert (
            value(
                m.fs.tb_distillate.properties_out[0].flow_mass_phase_comp["Liq", "TDS"]
            )
            == 1e-5
        )

        # Costing
        assert m.fs.costing.TIC.is_fixed()
        assert value(m.fs.costing.TIC) == 2
        assert m.fs.costing.heat_exchanger.material_factor_cost.is_fixed()
        assert value(m.fs.costing.heat_exchanger.material_factor_cost) == 5
        assert m.fs.costing.evaporator.material_factor_cost.is_fixed()
        assert value(m.fs.costing.evaporator.material_factor_cost) == 5
        assert m.fs.costing.compressor.unit_cost.is_fixed()
        assert value(m.fs.costing.compressor.unit_cost) == 7364

        # Temperature upper bounds
        assert value(m.fs.evaporator.properties_vapor[0].temperature.ub) == 348.15
        assert (
            value(m.fs.compressor.control_volume.properties_out[0].temperature.ub)
            == 450
        )

        # check degrees of freedom
        assert degrees_of_freedom(m) == 1

    @pytest.mark.component
    @pytest.mark.requires_idaes_solver
    def test_initialize_system(self, mvc_single_stage):
        m = mvc_single_stage

        initialize_system(m)

        # mass flows in evaporator
        assert value(
            m.fs.evaporator.properties_vapor[0].flow_mass_phase_comp["Vap", "H2O"]
        ) == pytest.approx(22.222, rel=1e-3)
        assert value(
            m.fs.evaporator.properties_vapor[0].flow_mass_phase_comp["Liq", "H2O"]
        ) == pytest.approx(1e-8, abs=1e-3)
        assert value(
            m.fs.evaporator.properties_brine[0].flow_mass_phase_comp["Liq", "H2O"]
        ) == pytest.approx(17.777, rel=1e-3)
        assert value(
            m.fs.evaporator.properties_brine[0].flow_mass_phase_comp["Liq", "TDS"]
        ) == pytest.approx(4.444, rel=1e-3)

        # evaporator pressure
        assert value(m.fs.evaporator.properties_vapor[0].pressure) == pytest.approx(
            26.231e3, rel=1e-3
        )

        # compressor pressure outlet
        assert value(m.fs.compressor.control_volume.properties_out[0].pressure) / value(
            m.fs.compressor.control_volume.properties_in[0].pressure
        ) == pytest.approx(1.6, rel=1e-3)

    @pytest.mark.component
    @pytest.mark.requires_idaes_solver
    def test_simulation_Q_ext(self, mvc_single_stage):
        m = mvc_single_stage

        scale_costs(m)
        fix_outlet_pressures(m)

        assert degrees_of_freedom(m) == 1

        m.fs.objective = Objective(expr=m.fs.Q_ext[0])
        results = solve(m, solver=solver, tee=False)
        assert_optimal_termination(results)

        # Check system metrics
        assert value(m.fs.costing.specific_energy_consumption) == pytest.approx(
            22.02, rel=1e-2
        )
        assert value(m.fs.costing.LCOW) == pytest.approx(23.47, rel=1e-2)

        # Check mass balance
        assert pytest.approx(
            value(m.fs.feed.outlet.flow_mass_phase_comp[0, "Liq", "H2O"]), rel=1e-3
        ) == value(m.fs.distillate.inlet.flow_mass_phase_comp[0, "Liq", "H2O"]) + value(
            m.fs.brine.inlet.flow_mass_phase_comp[0, "Liq", "H2O"]
        )
        assert pytest.approx(
            value(m.fs.feed.outlet.flow_mass_phase_comp[0, "Liq", "TDS"]), rel=1e-3
        ) == value(m.fs.distillate.inlet.flow_mass_phase_comp[0, "Liq", "TDS"]) + value(
            m.fs.brine.inlet.flow_mass_phase_comp[0, "Liq", "TDS"]
        )

    @pytest.mark.requires_idaes_solver
    @pytest.mark.component
    def test_optimization(self, mvc_single_stage):
        m = mvc_single_stage
        m.fs.Q_ext[0].fix(0)  # no longer want external heating in evaporator
        del m.fs.objective
        set_up_optimization(m)
        assert number_total_objectives(m) == 1
        assert degrees_of_freedom(m) == 4
        results = solve(m, solver=solver, tee=False)
        assert_optimal_termination(results)
        # Check decision variables
        assert value(m.fs.evaporator.properties_brine[0].temperature) == pytest.approx(
            348.15, rel=1e-2
        )
        assert value(m.fs.evaporator.properties_brine[0].pressure) == pytest.approx(
            32448.24, rel=1e-2
        )
        assert value(m.fs.hx_brine.area) == pytest.approx(173.99, rel=1e-2)
        assert value(m.fs.hx_distillate.area) == pytest.approx(206.31, rel=1e-2)
        assert value(m.fs.compressor.pressure_ratio) == pytest.approx(1.61, rel=1e-2)
        assert value(m.fs.evaporator.area) == pytest.approx(777.37, rel=1e-2)
        assert value(m.fs.evaporator.lmtd) == pytest.approx(22.59, rel=1e-2)

        # Check system metrics
        assert value(m.fs.costing.specific_energy_consumption) == pytest.approx(
            22.36, rel=1e-2
        )
        assert value(m.fs.costing.LCOW) == pytest.approx(4.52, rel=1e-2)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import os
import pytest
import tempfile
from watertap.examples.flowsheets.case_studies.electroNP import (
    multi_sweep,
)

sweep_list = []
for case_num in [1]:
    sweep_list.append(case_num)


@pytest.mark.parametrize("case_num", sweep_list)
@pytest.mark.integration
def test_multi_sweep(case_num, tmp_path):
    cwd = os.getcwd()
    os.chdir(tmp_path)
    nx = 1
    temp = tempfile.NamedTemporaryFile(delete=False)
    temp.close()
    global_results, sweep_params, m = multi_sweep.run_analysis(
        case_num,
        nx,
        interpolate_nan_outputs=False,
        output_filename=temp.name,
    )
    os.remove(temp.name)
    os.chdir(cwd)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import numpy as np


class DummyCOMM:

    rank = 0
    size = 1

    @staticmethod
    def Get_rank():
        return 0

    @staticmethod
    def Get_size():
        return 1

    @staticmethod
    def Barrier():
        pass

    @staticmethod
    def Bcast(array, root=0):
        pass

    @staticmethod
    def bcast(array, root=0):
        return array

    @staticmethod
    def allgather(value):
        return [value]

    @staticmethod
    def Gatherv(sendbuf=None, recvbuf=(None, None), root=0):
        assert len(recvbuf) == 2
        assert isinstance(recvbuf, tuple)
        receive_arr = recvbuf[0]
        receive_sizes = recvbuf[1]

        assert isinstance(receive_arr, (np.ndarray, list))
        assert len(receive_arr) == sum(receive_sizes)
        receive_arr[:] = sendbuf[:]



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest
from pyomo.environ import (
    ConcreteModel,
    Var,
    Expression,
    Constraint,
    TerminationCondition,
    SolverStatus,
    value,
)
from pyomo.network import Port
from idaes.core import (
    FlowsheetBlock,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
)
from watertap.unit_models.pressure_exchanger import PressureExchanger
import watertap.property_models.seawater_prop_pack as props
import watertap.property_models.seawater_ion_prop_pack as property_seawater_ions

from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_variables,
    number_total_constraints,
    number_unused_variables,
)
from idaes.core.solvers import get_solver
from idaes.core.util.testing import initialization_tester
from idaes.core.util.scaling import (
    calculate_scaling_factors,
    unscaled_variables_generator,
    unscaled_constraints_generator,
    badly_scaled_var_generator,
)

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


# # -----------------------------------------------------------------------------


@pytest.mark.unit
def test_config_no_mass_transfer():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = props.SeawaterParameterBlock()
    m.fs.unit = PressureExchanger(property_package=m.fs.properties)

    # check unit config arguments
    assert len(m.fs.unit.config) == 9

    assert not m.fs.unit.config.dynamic
    assert not m.fs.unit.config.has_holdup
    assert m.fs.unit.config.material_balance_type == MaterialBalanceType.useDefault
    assert m.fs.unit.config.energy_balance_type == EnergyBalanceType.none
    assert m.fs.unit.config.momentum_balance_type == MomentumBalanceType.pressureTotal
    assert m.fs.unit.config.property_package is m.fs.properties

    # verify no mass transfer
    assert not m.fs.unit.config.has_mass_transfer


@pytest.mark.unit
def test_config_mass_transfer():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = props.SeawaterParameterBlock()
    m.fs.unit = PressureExchanger(
        property_package=m.fs.properties, has_mass_transfer=True
    )
    # check mass_transfer is added
    assert m.fs.unit.config.has_mass_transfer


@pytest.mark.unit
def test_build(has_mass_transfer=False, extra_variables=0, extra_constraint=0):
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = props.SeawaterParameterBlock()
    m.fs.unit = PressureExchanger(
        property_package=m.fs.properties, has_mass_transfer=has_mass_transfer
    )

    # test ports and state variables
    port_lst = [
        "low_pressure_inlet",
        "low_pressure_outlet",
        "high_pressure_inlet",
        "high_pressure_outlet",
    ]
    port_vars_lst = ["flow_mass_phase_comp", "pressure", "temperature"]

    for port_str in port_lst:
        assert hasattr(m.fs.unit, port_str)
        port = getattr(m.fs.unit, port_str)
        assert len(port.vars) == 3
        assert isinstance(port, Port)
        for var_str in port_vars_lst:
            assert hasattr(port, var_str)
            var = getattr(port, var_str)
            assert isinstance(var, Var)

    # test unit variables
    assert hasattr(m.fs.unit, "efficiency_pressure_exchanger")
    assert isinstance(m.fs.unit.efficiency_pressure_exchanger, Var)
    if not has_mass_transfer:
        assert not hasattr(m.fs.unit, "mass_transfer_fraction_comp")
    else:
        assert isinstance(m.fs.unit.mass_transfer_fraction_comp, Var)

    # test unit constraints

    unit_cons_lst = [
        "eq_pressure_transfer",
        "eq_equal_flow_vol",
        "eq_equal_low_pressure",
    ]

    if has_mass_transfer:
        unit_cons_lst.append("eq_mass_transfer_from_high_pressure_side")
        unit_cons_lst.append("eq_mass_transfer_term")

    for c in unit_cons_lst:
        assert hasattr(m.fs.unit, c)
        con = getattr(m.fs.unit, c)
        assert isinstance(con, Constraint)

    # test control volumes, only terms directly used by pressure exchanger

    cv_list = ["low_pressure_side", "high_pressure_side"]
    cv_var_lst = ["deltaP"]
    if has_mass_transfer:
        cv_var_lst.append("mass_transfer_term")
    cv_exp_lst = ["work"]

    for cv_str in cv_list:
        assert hasattr(m.fs.unit, cv_str)
        cv = getattr(m.fs.unit, cv_str)
        for cv_var_str in cv_var_lst:
            cv_var = getattr(cv, cv_var_str)
            assert isinstance(cv_var, Var)
        for cv_exp_str in cv_exp_lst:
            cv_exp = getattr(cv, cv_exp_str)
            assert isinstance(cv_exp, Expression)

    # test state blocks, only terms directly used by pressure exchanger
    cv_stateblock_lst = ["properties_in", "properties_out"]
    stateblock_var_lst = ["pressure", "temperature"]
    stateblock_exp_lst = ["flow_vol"]
    for cv_str in cv_list:
        cv = getattr(m.fs.unit, cv_str)
        for cv_sb_str in cv_stateblock_lst:
            assert hasattr(cv, cv_sb_str)
            cv_sb = getattr(cv, cv_sb_str)
            for sb_var_str in stateblock_var_lst:
                assert hasattr(cv_sb[0], sb_var_str)
                sb_var = getattr(cv_sb[0], sb_var_str)
                assert isinstance(sb_var, Var)
            for sb_exp_str in stateblock_exp_lst:
                assert hasattr(cv_sb[0], sb_exp_str)
                sb_exp = getattr(cv_sb[0], sb_exp_str)
                assert isinstance(sb_exp, Expression)

    # test statistics
    assert number_variables(m.fs.unit) == 39 + extra_variables
    assert number_total_constraints(m.fs.unit) == 31 + extra_constraint
    assert number_unused_variables(m.fs.unit) == 0


@pytest.mark.unit
def test_build_with_mass_transfer():
    # 2 variables for mass transfer fractions (solute,solvent)
    # 4 extra vars for mass_transfer_term on LP and HP side, and solute, solvent
    # 4 extra constraints for mass transsfer from HP side (soute, solvent)
    test_build(has_mass_transfer=True, extra_variables=6, extra_constraint=4)


class TestPressureExchanger:
    @pytest.fixture(scope="class")
    def unit_frame(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = props.SeawaterParameterBlock()
        m.fs.unit = PressureExchanger(
            property_package=m.fs.properties, has_mass_transfer=True
        )

        # Specify inlet conditions
        temperature = 25 + 273.15
        flow_vol = 1e-3
        lowP_mass_frac_TDS = 0.035
        lowP_pressure = 101325
        highP_mass_frac_TDS = 0.07
        highP_pressure = 50e5

        solute_transfer = 0.05
        solvent_transfer = 0.1
        m.fs.unit.mass_transfer_fraction_comp[0, "H2O"].fix(solvent_transfer)
        m.fs.unit.mass_transfer_fraction_comp[0, "TDS"].fix(solute_transfer)

        m.fs.unit.low_pressure_side.properties_in[0].flow_vol_phase["Liq"].fix(flow_vol)
        m.fs.unit.low_pressure_side.properties_in[0].mass_frac_phase_comp[
            "Liq", "TDS"
        ].fix(lowP_mass_frac_TDS)

        m.fs.unit.low_pressure_side.properties_in[0].pressure.fix(lowP_pressure)
        m.fs.unit.low_pressure_side.properties_in[0].temperature.fix(temperature)

        m.fs.unit.high_pressure_side.properties_in[0].flow_vol_phase["Liq"].fix(
            flow_vol
        )
        m.fs.unit.high_pressure_side.properties_in[0].mass_frac_phase_comp[
            "Liq", "TDS"
        ].fix(highP_mass_frac_TDS)
        m.fs.unit.high_pressure_side.properties_in[0].pressure.fix(highP_pressure)
        m.fs.unit.high_pressure_side.properties_in[0].temperature.fix(temperature)

        # solve inlet conditions and only fix state variables (i.e. unfix flow_vol and mass_frac_phase)

        results = solver.solve(m.fs.unit.high_pressure_side.properties_in[0])
        assert results.solver.termination_condition == TerminationCondition.optimal
        m.fs.unit.high_pressure_side.properties_in[0].flow_mass_phase_comp[
            "Liq", "TDS"
        ].fix()
        m.fs.unit.high_pressure_side.properties_in[0].flow_vol_phase["Liq"].unfix()
        m.fs.unit.high_pressure_side.properties_in[0].mass_frac_phase_comp[
            "Liq", "TDS"
        ].unfix()

        results = solver.solve(m.fs.unit.low_pressure_side.properties_in[0])
        assert results.solver.termination_condition == TerminationCondition.optimal
        m.fs.unit.low_pressure_side.properties_in[0].flow_mass_phase_comp[
            "Liq", "H2O"
        ].fix()
        m.fs.unit.low_pressure_side.properties_in[0].flow_mass_phase_comp[
            "Liq", "TDS"
        ].fix()
        m.fs.unit.low_pressure_side.properties_in[0].flow_vol_phase["Liq"].unfix()
        m.fs.unit.low_pressure_side.properties_in[0].mass_frac_phase_comp[
            "Liq", "TDS"
        ].unfix()

        # Specify unit
        efficiency = 0.95
        m.fs.unit.efficiency_pressure_exchanger.fix(efficiency)
        return m

    @pytest.mark.unit
    def test_dof(self, unit_frame):
        assert degrees_of_freedom(unit_frame) == 0

    @pytest.mark.unit
    def test_calculate_scaling(self, unit_frame):
        m = unit_frame
        calculate_scaling_factors(m)

        # check that all variables have scaling factors
        unscaled_var_list = list(
            unscaled_variables_generator(m.fs.unit, include_fixed=True)
        )
        assert len(unscaled_var_list) == 0
        # check that all constraints have been scaled
        unscaled_constraint_list = list(unscaled_constraints_generator(m))
        assert len(unscaled_constraint_list) == 0

    @pytest.mark.component
    def test_initialize(self, unit_frame):
        m = unit_frame
        initialization_tester(unit_frame)

    @pytest.mark.component
    def test_var_scaling(self, unit_frame):
        m = unit_frame
        badly_scaled_var_lst = list(badly_scaled_var_generator(m))
        assert badly_scaled_var_lst == []

    @pytest.mark.component
    def test_solve(self, unit_frame):
        m = unit_frame
        results = solver.solve(m)

        # Check for optimal solution
        assert results.solver.termination_condition == TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.component
    def test_solution(self, unit_frame):
        m = unit_frame
        assert pytest.approx(0.9877, rel=1e-3) == value(
            m.fs.unit.low_pressure_inlet.flow_mass_phase_comp[0, "Liq", "H2O"]
        )
        assert pytest.approx(3.582e-2, rel=1e-3) == value(
            m.fs.unit.low_pressure_inlet.flow_mass_phase_comp[0, "Liq", "TDS"]
        )
        assert pytest.approx(4.755e6, rel=1e-3) == value(
            m.fs.unit.low_pressure_outlet.pressure[0]
        )
        assert pytest.approx(0.9877 * 1.1, rel=1e-3) == value(
            m.fs.unit.high_pressure_inlet.flow_mass_phase_comp[0, "Liq", "H2O"]
        )
        assert pytest.approx(7.352e-2, rel=1e-3) == value(
            m.fs.unit.high_pressure_inlet.flow_mass_phase_comp[0, "Liq", "TDS"]
        )
        assert pytest.approx(4.654e6, rel=1e-3) == value(
            m.fs.unit.low_pressure_side.deltaP[0]
        )
        assert pytest.approx(4.654e3, rel=1e-3) == value(
            m.fs.unit.low_pressure_side.work[0]
        )
        assert pytest.approx(-4.899e6, rel=1e-3) == value(
            m.fs.unit.high_pressure_side.deltaP[0]
        )
        assert pytest.approx(-5.436e3, rel=1e-3) == value(
            m.fs.unit.high_pressure_side.work[0]
        )

        # testing solvent transfer
        assert pytest.approx(-0.9877 * 1.1 * 0.1, rel=1e-3) == value(
            m.fs.unit.high_pressure_side.mass_transfer_term[0, "Liq", "H2O"]
        )
        assert pytest.approx(0.9778, rel=1e-3) == value(
            m.fs.unit.high_pressure_outlet.flow_mass_phase_comp[0, "Liq", "H2O"]
        )
        assert pytest.approx(0.9877 + 0.9877 * 1.1 * 0.1, rel=1e-3) == value(
            m.fs.unit.low_pressure_outlet.flow_mass_phase_comp[0, "Liq", "H2O"]
        )
        # testing solute transfer
        assert pytest.approx(-7.352e-2 * 0.05, rel=1e-3) == value(
            m.fs.unit.high_pressure_side.mass_transfer_term[0, "Liq", "TDS"]
        )
        assert pytest.approx(7.352e-2 - 7.352e-2 * 0.05, rel=1e-3) == value(
            m.fs.unit.high_pressure_outlet.flow_mass_phase_comp[0, "Liq", "TDS"]
        )
        assert pytest.approx(3.582e-2 + 7.352e-2 * 0.05, rel=1e-3) == value(
            m.fs.unit.low_pressure_outlet.flow_mass_phase_comp[0, "Liq", "TDS"]
        )

    @pytest.mark.unit
    def test_report(self, unit_frame):
        unit_frame.fs.unit.report()


class TestPressureExchanger_with_ion_prop_pack:
    @pytest.fixture(scope="class")
    def unit_frame(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = property_seawater_ions.PropParameterBlock()
        m.fs.unit = PressureExchanger(
            property_package=m.fs.properties, has_mass_transfer=True
        )

        # Specify inlet conditions
        temperature = 25 + 273.15

        lowP_pressure = 101325
        highP_mass_frac_TDS = 0.07
        highP_pressure = 50e5

        solvent_transfer = 0.05
        solute_transfer = {"Na": 0.1, "Ca": 0.11, "Mg": 0.12, "SO4": 0.13, "Cl": 0.14}

        feed_flow_mass = 1
        lowP_mass_frac = {
            "Na": 11122e-6,
            "Ca": 382e-6,
            "Mg": 1394e-6,
            "SO4": 2136e-6,
            "Cl": 20300e-6,
        }

        highP_mass_frac = {
            "Na": 2 * 11122e-6,
            "Ca": 2 * 382e-6,
            "Mg": 2 * 1394e-6,
            "SO4": 2 * 2136e-6,
            "Cl": 2 * 20300e-6,
        }

        # Specify unit
        efficiency = 0.95
        m.fs.unit.efficiency_pressure_exchanger.fix(efficiency)

        m.fs.unit.mass_transfer_fraction_comp[0, "H2O"].fix(solvent_transfer)
        for s in solute_transfer:
            m.fs.unit.mass_transfer_fraction_comp[0, s].fix(solute_transfer[s])

        m.fs.unit.low_pressure_side.properties_in[0].flow_mass_phase_comp[
            "Liq", "H2O"
        ].fix(feed_flow_mass * (1 - sum(x for x in lowP_mass_frac.values())))
        for s in lowP_mass_frac:
            m.fs.unit.low_pressure_side.properties_in[0].flow_mass_phase_comp[
                "Liq", s
            ].fix(feed_flow_mass * lowP_mass_frac[s])
        m.fs.unit.low_pressure_side.properties_in[0].pressure.fix(lowP_pressure)
        m.fs.unit.low_pressure_side.properties_in[0].temperature.fix(temperature)

        ### NOTE THE PRESSURE EXHCANGER EXPECTS EQUAL VOLUMETRIC FLOW RATES
        # so one of mass fraction (here we unfix H2O mass fraction)
        # s on high pressure or low pressure sides
        # should be left unfixed, as other wise you are fixing all the mass fractions,
        # and forcing equal flow, which leaves no degrees of freedom
        # to satisfy the equal flow constriants, which will lead to
        # a -1 DOF error and should not occur in practice where typicly at least one of
        # the inlet mass flow rates is colaculated.

        m.fs.unit.high_pressure_side.properties_in[0].flow_mass_phase_comp[
            "Liq", "H2O"
        ] = feed_flow_mass * (1 - sum(x for x in highP_mass_frac.values()))

        for s in highP_mass_frac:
            m.fs.unit.high_pressure_side.properties_in[0].flow_mass_phase_comp[
                "Liq", s
            ].fix(feed_flow_mass * highP_mass_frac[s])

        m.fs.unit.high_pressure_side.properties_in[0].pressure.fix(highP_pressure)
        m.fs.unit.high_pressure_side.properties_in[0].temperature.fix(temperature)

        return m

    @pytest.mark.unit
    def test_dof(self, unit_frame):
        assert degrees_of_freedom(unit_frame) == 0

    @pytest.mark.unit
    def test_calculate_scaling(self, unit_frame):
        m = unit_frame
        calculate_scaling_factors(m)

        # check that all variables have scaling factors
        unscaled_var_list = list(
            unscaled_variables_generator(m.fs.unit, include_fixed=True)
        )
        assert len(unscaled_var_list) == 0
        # check that all constraints have been scaled
        unscaled_constraint_list = list(unscaled_constraints_generator(m))
        assert len(unscaled_constraint_list) == 0

    @pytest.mark.component
    def test_initialize(self, unit_frame):
        m = unit_frame
        initialization_tester(unit_frame)

    @pytest.mark.component
    def test_var_scaling(self, unit_frame):
        m = unit_frame
        badly_scaled_var_lst = list(badly_scaled_var_generator(m))
        assert badly_scaled_var_lst == []

    @pytest.mark.component
    def test_solve(self, unit_frame):
        m = unit_frame
        results = solver.solve(m)

        # Check for optimal solution
        assert results.solver.termination_condition == TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.component
    def test_solution(self, unit_frame):
        m = unit_frame

        solvent_transfer = 0.05
        solute_transfer = {"Na": 0.1, "Ca": 0.11, "Mg": 0.12, "SO4": 0.13, "Cl": 0.14}

        feed_flow_mass = 1
        lowP_mass_frac = {
            "Na": 11122e-6,
            "Ca": 382e-6,
            "Mg": 1394e-6,
            "SO4": 2136e-6,
            "Cl": 20300e-6,
        }
        highP_mass_frac = {
            "Na": 2 * 11122e-6,
            "Ca": 2 * 382e-6,
            "Mg": 2 * 1394e-6,
            "SO4": 2 * 2136e-6,
            "Cl": 2 * 20300e-6,
        }

        assert pytest.approx(
            feed_flow_mass * (1 - sum(x for x in lowP_mass_frac.values())), rel=1e-3
        ) == value(m.fs.unit.low_pressure_inlet.flow_mass_phase_comp[0, "Liq", "H2O"])

        for s in lowP_mass_frac:
            assert pytest.approx(feed_flow_mass * lowP_mass_frac[s], rel=1e-3) == value(
                m.fs.unit.low_pressure_inlet.flow_mass_phase_comp[0, "Liq", s]
            )

        assert pytest.approx(4.755e6, rel=1e-3) == value(
            m.fs.unit.low_pressure_outlet.pressure[0]
        )

        assert pytest.approx(0.9876, rel=1e-3) == value(
            m.fs.unit.high_pressure_inlet.flow_mass_phase_comp[0, "Liq", "H2O"]
        )

        for s in highP_mass_frac:
            assert pytest.approx(
                feed_flow_mass * highP_mass_frac[s], rel=1e-3
            ) == value(m.fs.unit.high_pressure_inlet.flow_mass_phase_comp[0, "Liq", s])

        assert pytest.approx(4.654e6, rel=1e-3) == value(
            m.fs.unit.low_pressure_side.deltaP[0]
        )
        assert pytest.approx(4.654e3, rel=1e-3) == value(
            m.fs.unit.low_pressure_side.work[0]
        )
        assert pytest.approx(-4.899e6, rel=1e-3) == value(
            m.fs.unit.high_pressure_side.deltaP[0]
        )
        assert pytest.approx(-5.184e3, rel=1e-3) == value(
            m.fs.unit.high_pressure_side.work[0]
        )

        # testing solvent transfer
        assert pytest.approx(-0.9876 * solvent_transfer, rel=1e-3) == value(
            m.fs.unit.high_pressure_side.mass_transfer_term[0, "Liq", "H2O"]
        )
        assert pytest.approx(0.9876 - 0.9876 * solvent_transfer, rel=1e-3) == value(
            m.fs.unit.high_pressure_outlet.flow_mass_phase_comp[0, "Liq", "H2O"]
        )
        assert pytest.approx(
            feed_flow_mass * (1 - sum(x for x in lowP_mass_frac.values()))
            + 0.9876 * solvent_transfer,
            rel=1e-3,
        ) == value(m.fs.unit.low_pressure_outlet.flow_mass_phase_comp[0, "Liq", "H2O"])
        # testing solute transfer
        for s in highP_mass_frac:
            assert pytest.approx(
                -feed_flow_mass * highP_mass_frac[s] * solute_transfer[s], rel=1e-3
            ) == value(m.fs.unit.high_pressure_side.mass_transfer_term[0, "Liq", s])
            assert pytest.approx(
                feed_flow_mass * highP_mass_frac[s]
                - feed_flow_mass * highP_mass_frac[s] * solute_transfer[s],
                rel=1e-3,
            ) == value(m.fs.unit.high_pressure_outlet.flow_mass_phase_comp[0, "Liq", s])
            assert pytest.approx(
                feed_flow_mass * lowP_mass_frac[s]
                + feed_flow_mass * highP_mass_frac[s] * solute_transfer[s],
                rel=1e-3,
            ) == value(m.fs.unit.low_pressure_outlet.flow_mass_phase_comp[0, "Liq", s])

    @pytest.mark.unit
    def test_report(self, unit_frame):
        unit_frame.fs.unit.report()



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest
from pyomo.environ import (
    ConcreteModel,
    Var,
    value,
)
from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.models.unit_models import Product, Feed
from pyomo.util.check_units import assert_units_consistent
import watertap.property_models.NaCl_prop_pack as props
from watertap.unit_models.reverse_osmosis_0D import ReverseOsmosis0D
from watertap.unit_models.osmotically_assisted_reverse_osmosis_0D import (
    OsmoticallyAssistedReverseOsmosis0D,
)

from watertap.unit_models.pressure_changer import Pump, EnergyRecoveryDevice
from watertap.examples.flowsheets.oaro.oaro import (
    main,
    build,
    set_operating_conditions,
    initialize_system,
    solve,
    ERDtype,
)

solver = get_solver()

# -----------------------------------------------------------------------------
class TestOAROwithTurbine:
    @pytest.fixture(scope="class")
    def system_frame(self):
        m = build(erd_type=ERDtype.pump_as_turbine)

        return m

    @pytest.mark.unit
    def test_build(self, system_frame):
        m = system_frame
        # model set up
        assert isinstance(m, ConcreteModel)
        assert isinstance(m.fs, FlowsheetBlock)
        assert isinstance(m.fs.properties, props.NaClParameterBlock)

        # unit models
        fs = m.fs
        assert isinstance(fs.feed, Feed)
        assert isinstance(fs.P1, Pump)
        assert isinstance(fs.P2, Pump)
        assert isinstance(fs.P3, Pump)
        assert isinstance(fs.ERD1, EnergyRecoveryDevice)
        assert isinstance(fs.ERD2, EnergyRecoveryDevice)
        assert isinstance(fs.OARO, OsmoticallyAssistedReverseOsmosis0D)
        assert isinstance(fs.RO, ReverseOsmosis0D)
        assert isinstance(fs.product, Product)
        assert isinstance(fs.disposal, Product)

        # additional variables
        assert isinstance(fs.volumetric_recovery, Var)
        assert isinstance(fs.water_recovery, Var)

        # arcs
        arc_dict = {
            fs.s01: (fs.feed.outlet, fs.P1.inlet),
            fs.s02: (fs.P1.outlet, fs.OARO.feed_inlet),
            fs.s03: (fs.OARO.feed_outlet, fs.ERD1.inlet),
            fs.s04: (fs.ERD1.outlet, fs.disposal.inlet),
            fs.s05: (fs.OARO.permeate_outlet, fs.P2.inlet),
            fs.s06: (fs.P2.outlet, fs.RO.inlet),
            fs.s07: (fs.RO.permeate, fs.product.inlet),
            fs.s08: (fs.RO.retentate, fs.ERD2.inlet),
            fs.s09: (fs.ERD2.outlet, fs.P3.inlet),
            fs.s10: (fs.P3.outlet, fs.OARO.permeate_inlet),
        }
        for arc, port_tpl in arc_dict.items():
            assert arc.source is port_tpl[0]
            assert arc.destination is port_tpl[1]

    @pytest.mark.component
    def test_units(self, system_frame):
        assert_units_consistent(system_frame)

    @pytest.mark.component
    def test_set_operating_conditions(self, system_frame):
        m = system_frame
        set_operating_conditions(m)
        assert degrees_of_freedom(m) == 0

    @pytest.mark.component
    def test_initialize_system(self, system_frame):
        m = system_frame
        initialize_system(m, solver=solver)
        assert degrees_of_freedom(m) == 0

    @pytest.mark.component
    def test_solution(self, system_frame):
        m = system_frame
        solve(m, solver=solver)
        fs = m.fs
        assert pytest.approx(2.8614e-5, rel=1e-5) == value(
            fs.product.flow_mass_phase_comp[0, "Liq", "NaCl"]
        )
        assert pytest.approx(0.253447, rel=1e-5) == value(fs.water_recovery)

    @pytest.mark.component
    def test_config_error(self, system_frame):
        with pytest.raises(Exception):
            build(erd_type="not_a_configuration")

    @pytest.mark.component
    def test_main(self):
        main(raise_on_failure=True)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

# Import Pyomo libraries
from pyomo.environ import (
    Var,
    NonNegativeReals,
    value,
)

from idaes.core import declare_process_block_class, FlowDirection
from idaes.core.util import scaling as iscale
from idaes.core.util.misc import add_object_reference
from watertap.core import (
    MembraneChannel0DBlock,
    MassTransferCoefficient,
    PressureChangeType,
)
from watertap.core.membrane_channel0d import CONFIG_Template
from watertap.unit_models.osmotically_assisted_reverse_osmosis_base import (
    OsmoticallyAssistedReverseOsmosisBaseData,
    _add_has_full_reporting,
)


__author__ = "Adam Atia, Chenyu Wang"


@declare_process_block_class("OsmoticallyAssistedReverseOsmosis0D")
class OsmoticallyAssistedReverseOsmosisData(OsmoticallyAssistedReverseOsmosisBaseData):
    """
    Standard OARO Unit Model Class:
    - zero dimensional model
    - steady state only
    - single liquid phase only
    """

    CONFIG = CONFIG_Template()

    _add_has_full_reporting(CONFIG)

    def _add_membrane_channels_and_geometry(self):

        # Build membrane channel control volume
        channel_kwargs = {
            "dynamic": False,
            "has_holdup": False,
            "property_package": self.config.property_package,
            "property_package_args": self.config.property_package_args,
        }
        self.feed_side = MembraneChannel0DBlock(**channel_kwargs)
        self.permeate_side = MembraneChannel0DBlock(**channel_kwargs)

        if (self.config.pressure_change_type != PressureChangeType.fixed_per_stage) or (
            self.config.mass_transfer_coefficient == MassTransferCoefficient.calculated
        ):
            self._add_length_and_width()
            add_geometry_kwargs = {
                "length_var": self.length,
                "width_var": self.width,
            }
            self.feed_side.add_geometry(
                flow_direction=FlowDirection.forward, **add_geometry_kwargs
            )
            self.permeate_side.add_geometry(
                flow_direction=FlowDirection.backward, **add_geometry_kwargs
            )
            self._add_area(include_constraint=True)
        else:
            add_geometry_kwargs = {
                "length_var": None,
                "width_var": None,
            }
            self.feed_side.add_geometry(
                flow_direction=FlowDirection.forward, **add_geometry_kwargs
            )
            self.permeate_side.add_geometry(
                flow_direction=FlowDirection.backward, **add_geometry_kwargs
            )
            self._add_area(include_constraint=False)

    def _add_deltaP(self, side):
        mem_side = self.component(side)
        add_object_reference(mem_side, "deltaP_stage", mem_side.deltaP)

    def _add_mass_transfer(self):

        units_meta = self.config.property_package.get_metadata().get_derived_units

        # mass transfer
        def mass_transfer_phase_comp_initialize(b, t, p, j):
            return value(
                self.feed_side.properties_in[t].get_material_flow_terms("Liq", j)
                * self.recovery_mass_phase_comp[t, "Liq", j]
            )

        self.mass_transfer_phase_comp = Var(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            initialize=mass_transfer_phase_comp_initialize,
            bounds=(0.0, 1e6),
            domain=NonNegativeReals,
            units=units_meta("mass") * units_meta("time") ** -1,
            doc="Mass transfer to permeate",
        )

        @self.Constraint(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            doc="Mass transfer term",
        )
        def eq_mass_transfer_term(self, t, p, j):
            return (
                self.mass_transfer_phase_comp[t, p, j]
                == -self.feed_side.mass_transfer_term[t, p, j]
            )

        # Feed and permeate-side connection
        @self.Constraint(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            doc="Mass transfer from feed to permeate",
        )
        def eq_connect_mass_transfer(b, t, p, j):
            return (
                b.permeate_side.mass_transfer_term[t, p, j]
                == -b.feed_side.mass_transfer_term[t, p, j]
            )

        @self.Constraint(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            doc="Permeate production",
        )
        def eq_permeate_production(b, t, p, j):
            return (
                b.permeate_side.mass_transfer_term[t, p, j]
                == b.area * b.flux_mass_phase_comp_avg[t, p, j]
            )

    def calculate_scaling_factors(self):

        super().calculate_scaling_factors()

        for (t, p, j), v in self.mass_transfer_phase_comp.items():
            sf = iscale.get_scaling_factor(
                self.feed_side.properties_in[t].get_material_flow_terms(p, j)
            )
            if iscale.get_scaling_factor(v) is None:
                iscale.set_scaling_factor(v, sf)
            v = self.feed_side.mass_transfer_term[t, p, j]
            if iscale.get_scaling_factor(v) is None:
                iscale.set_scaling_factor(v, sf)
            v = self.permeate_side.mass_transfer_term[t, p, j]
            if iscale.get_scaling_factor(v) is None:
                iscale.set_scaling_factor(v, sf)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import os
import idaes.logger as idaeslog

from pyomo.environ import (
    ConcreteModel,
    units as pyunits,
    Expression,
    value,
    TransformationFactory,
)

from pyomo.network import Arc, SequentialDecomposition
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.models.unit_models import Product
import idaes.core.util.scaling as iscale
from idaes.core import UnitModelCostingBlock

from watertap.core.util.initialization import assert_degrees_of_freedom, check_solve

from watertap.core.wt_database import Database
import watertap.core.zero_order_properties as prop_ZO
from watertap.unit_models.zero_order import (
    FeedZO,
    MicrobialBatteryZO,
    PumpElectricityZO,
)
from watertap.costing.zero_order_costing import ZeroOrderCosting

# Set up logger
_log = idaeslog.getLogger(__name__)


def main():
    m = build()

    set_operating_conditions(m)
    assert_degrees_of_freedom(m, 0)
    assert_units_consistent(m)

    initialize_system(m)

    results = solve(m, checkpoint="solve flowsheet after initializing system")
    display_results(m)

    add_costing(m)
    m.fs.costing.initialize()

    assert_degrees_of_freedom(m, 0)
    assert_units_consistent(m)
    results = solve(m, checkpoint="solve flowsheet after costing")
    display_costing(m)

    return m, results


def build():
    # flowsheet set up
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.prop = prop_ZO.WaterParameterBlock(
        solute_list=[
            "arsenic",
            "uranium",
            "nitrate",
            "phosphates",
            "iron",
            "filtration_media",
        ]
    )

    # unit models
    # feed
    m.fs.feed = FeedZO(property_package=m.fs.prop)

    # pump
    m.fs.pump = PumpElectricityZO(property_package=m.fs.prop, database=m.db)

    # microbial battery
    m.fs.micbatt = MicrobialBatteryZO(property_package=m.fs.prop, database=m.db)

    # product streams
    m.fs.filtered_water = Product(property_package=m.fs.prop)
    m.fs.byproduct = Product(property_package=m.fs.prop)

    # connections
    m.fs.s01 = Arc(source=m.fs.feed.outlet, destination=m.fs.pump.inlet)
    m.fs.s02 = Arc(source=m.fs.pump.outlet, destination=m.fs.micbatt.inlet)
    m.fs.s03 = Arc(source=m.fs.micbatt.treated, destination=m.fs.filtered_water.inlet)
    m.fs.s04 = Arc(source=m.fs.micbatt.byproduct, destination=m.fs.byproduct.inlet)

    # expand arcs
    TransformationFactory("network.expand_arcs").apply_to(m)

    # scaling
    iscale.calculate_scaling_factors(m)

    return m


def set_operating_conditions(m):
    # ---specifications---
    # feed
    flow_vol = 1 * pyunits.m**3 / pyunits.day
    conc_mass_arsenic = 0.04 * pyunits.mg / pyunits.L
    conc_mass_uranium = 0.06 * pyunits.mg / pyunits.L
    conc_mass_nitrate = 10 * pyunits.mg / pyunits.L
    conc_mass_phosphates = 0.1 * pyunits.mg / pyunits.L
    conc_mass_iron = 0.5 * pyunits.mg / pyunits.L
    conc_mass_filtration_media = 0.1 * pyunits.mg / pyunits.L

    m.fs.feed.flow_vol[0].fix(flow_vol)
    m.fs.feed.conc_mass_comp[0, "arsenic"].fix(conc_mass_arsenic)
    m.fs.feed.conc_mass_comp[0, "uranium"].fix(conc_mass_uranium)
    m.fs.feed.conc_mass_comp[0, "nitrate"].fix(conc_mass_nitrate)
    m.fs.feed.conc_mass_comp[0, "phosphates"].fix(conc_mass_phosphates)
    m.fs.feed.conc_mass_comp[0, "iron"].fix(conc_mass_iron)
    m.fs.feed.conc_mass_comp[0, "filtration_media"].fix(conc_mass_filtration_media)

    solve(m.fs.feed, checkpoint="solve feed block")

    # pump
    m.fs.pump.load_parameters_from_database(use_default_removal=True)
    m.fs.pump.lift_height.fix(30)  # head of 30 m to pump groundwater

    # microbial battery
    m.fs.micbatt.load_parameters_from_database(use_default_removal=True)


def initialize_system(m):
    seq = SequentialDecomposition()
    seq.options.tear_set = []
    seq.options.iterLim = 1
    seq.run(m, lambda u: u.initialize())


def solve(blk, solver=None, checkpoint=None, tee=False, fail_flag=True):
    if solver is None:
        solver = get_solver()
    results = solver.solve(blk, tee=tee)
    check_solve(results, checkpoint=checkpoint, logger=_log, fail_flag=fail_flag)
    return results


def display_results(m):
    print("\n+++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
    print("++++++++++++++++++++ DISPLAY RESULTS ++++++++++++++++++++")
    print("+++++++++++++++++++++++++++++++++++++++++++++++++++++++++")

    unit_list = ["pump", "micbatt"]
    for u in unit_list:
        m.fs.component(u).report()

    print("\n---------- Feed volumetric flowrate ----------")
    feed_vol_flow = value(
        pyunits.convert(m.fs.feed.flow_vol[0], to_units=pyunits.m**3 / pyunits.day)
    )
    print(f"{feed_vol_flow} m^3/day")

    print("\n---------- Feed water concentrations ----------")
    as_feed_conc = value(
        pyunits.convert(
            m.fs.feed.conc_mass_comp[0, "arsenic"], to_units=pyunits.mg / pyunits.L
        )
    )
    print(f"Arsenic: {as_feed_conc:.6f} mg/L")

    u_feed_conc = value(
        pyunits.convert(
            m.fs.feed.conc_mass_comp[0, "uranium"], to_units=pyunits.mg / pyunits.L
        )
    )
    print(f"Uranium: {u_feed_conc:.6f} mg/L")

    nitrates_feed_conc = value(
        pyunits.convert(
            m.fs.feed.conc_mass_comp[0, "nitrate"], to_units=pyunits.mg / pyunits.L
        )
    )
    print(f"Nitrates: {nitrates_feed_conc:.6f} mg/L")

    phosphates_feed_conc = value(
        pyunits.convert(
            m.fs.feed.conc_mass_comp[0, "phosphates"], to_units=pyunits.mg / pyunits.L
        )
    )
    print(f"Phosphates: {phosphates_feed_conc:.6f} mg/L")

    iron_feed_conc = value(
        pyunits.convert(
            m.fs.feed.conc_mass_comp[0, "iron"], to_units=pyunits.mg / pyunits.L
        )
    )
    print(f"Iron: {iron_feed_conc:.6f} mg/L")

    media_feed_conc = value(
        pyunits.convert(
            m.fs.feed.conc_mass_comp[0, "filtration_media"],
            to_units=pyunits.mg / pyunits.L,
        )
    )
    print(f"Filtration media: {media_feed_conc:.6f} mg/L")

    print("\n---------- Treated water flowrate ----------")
    treated_vol_flow = value(
        pyunits.convert(
            m.fs.filtered_water.properties[0].flow_vol,
            to_units=pyunits.m**3 / pyunits.day,
        )
    )
    print(f"{treated_vol_flow} m^3/day")

    print("\n---------- Treated water concentrations ----------")
    as_treated_conc = value(
        pyunits.convert(
            m.fs.micbatt.properties_treated[0].conc_mass_comp["arsenic"],
            to_units=pyunits.mg / pyunits.L,
        )
    )
    print(f"Arsenic: {as_treated_conc:.6f} mg/L")

    u_treated_conc = value(
        pyunits.convert(
            m.fs.micbatt.properties_treated[0].conc_mass_comp["uranium"],
            to_units=pyunits.mg / pyunits.L,
        )
    )
    print(f"Uranium: {u_treated_conc:.6f} mg/L")

    nitrates_treated_conc = value(
        pyunits.convert(
            m.fs.micbatt.properties_treated[0].conc_mass_comp["nitrate"],
            to_units=pyunits.mg / pyunits.L,
        )
    )
    print(f"Nitrates: {nitrates_treated_conc:.6f} mg/L")

    phosphates_treated_conc = value(
        pyunits.convert(
            m.fs.micbatt.properties_treated[0].conc_mass_comp["phosphates"],
            to_units=pyunits.mg / pyunits.L,
        )
    )
    print(f"Phosphates: {phosphates_treated_conc:.6f} mg/L")

    iron_treated_conc = value(
        pyunits.convert(
            m.fs.micbatt.properties_treated[0].conc_mass_comp["iron"],
            to_units=pyunits.mg / pyunits.L,
        )
    )
    print(f"Iron: {iron_treated_conc:.6f} mg/L")

    print("\n---------- Byproduct mass flow rates ----------")
    as_byproduct_mass_flow = value(
        pyunits.convert(
            m.fs.micbatt.byproduct.flow_mass_comp[0, "arsenic"],
            to_units=pyunits.mg / pyunits.day,
        )
    )
    print(f"Arsenic: {as_byproduct_mass_flow:.2f} mg/day")

    u_byproduct_mass_flow = value(
        pyunits.convert(
            m.fs.micbatt.byproduct.flow_mass_comp[0, "uranium"],
            to_units=pyunits.mg / pyunits.day,
        )
    )
    print(f"Uranium: {u_byproduct_mass_flow:.2f} mg/day")

    n_byproduct_mass_flow = value(
        pyunits.convert(
            m.fs.micbatt.byproduct.flow_mass_comp[0, "nitrate"],
            to_units=pyunits.mg / pyunits.day,
        )
    )
    print(f"Nitrates: {n_byproduct_mass_flow:.2f} mg/day")

    p_byproduct_mass_flow = value(
        pyunits.convert(
            m.fs.micbatt.byproduct.flow_mass_comp[0, "phosphates"],
            to_units=pyunits.mg / pyunits.day,
        )
    )
    print(f"Phosphates: {p_byproduct_mass_flow:.2f} mg/day")

    iron_byproduct_mass_flow = value(
        pyunits.convert(
            m.fs.micbatt.byproduct.flow_mass_comp[0, "iron"],
            to_units=pyunits.mg / pyunits.day,
        )
    )
    print(f"Iron: {iron_byproduct_mass_flow:.2f} mg/day")

    media_byproduct_mass_flow = value(
        pyunits.convert(
            m.fs.micbatt.byproduct.flow_mass_comp[0, "filtration_media"],
            to_units=pyunits.mg / pyunits.day,
        )
    )
    print(f"Filtration media: {media_byproduct_mass_flow:.2f} mg/day")

    print("\n+++++++++++++++++++++++++++++++++++++++++++++++++++++++++")


def add_costing(m):
    source_file = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "groundwater_treatment_case_study.yaml",
    )

    m.fs.costing = ZeroOrderCosting(case_study_definition=source_file)

    costing_kwargs = {"flowsheet_costing_block": m.fs.costing}
    m.fs.pump.costing = UnitModelCostingBlock(**costing_kwargs)
    m.fs.micbatt.costing = UnitModelCostingBlock(**costing_kwargs)

    m.fs.costing.cost_process()

    m.fs.costing.add_electricity_intensity(m.fs.feed.properties[0].flow_vol)
    m.fs.costing.add_LCOW(m.fs.feed.properties[0].flow_vol)

    m.fs.costing.annual_water_inlet = Expression(
        expr=m.fs.costing.utilization_factor
        * pyunits.convert(
            m.fs.feed.properties[0].flow_vol,
            to_units=pyunits.m**3 / m.fs.costing.base_period,
        )
    )

    m.fs.costing.LCOT = Expression(
        expr=(
            m.fs.costing.total_capital_cost * m.fs.costing.capital_recovery_factor
            + m.fs.costing.total_operating_cost
        )
        / (
            pyunits.convert(
                m.fs.feed.flow_vol[0],
                to_units=pyunits.m**3 / m.fs.costing.base_period,
            )
            * m.fs.costing.utilization_factor
        ),
        doc="Levelized Cost of Treatment (with respect to feed flowrate)",
    )


def display_costing(m):
    print("\n+++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
    print("++++++++++++++++++++ DISPLAY COSTING ++++++++++++++++++++")
    print("+++++++++++++++++++++++++++++++++++++++++++++++++++++++++")

    print("\n---------- Levelized costs ----------")
    LCOT = value(
        pyunits.convert(m.fs.costing.LCOT, to_units=pyunits.USD_2020 / pyunits.m**3)
    )
    print(f"Levelized cost of treatment: {LCOT:.2f} $/m^3")

    LCOW = value(
        pyunits.convert(m.fs.costing.LCOW, to_units=pyunits.USD_2020 / pyunits.m**3)
    )
    print(f"Levelized cost of water: {LCOW:.2f} $/m^3")

    print("\n------------- Capital costs -------------")
    DCC_normalized = value(
        pyunits.convert(
            (m.fs.pump.costing.capital_cost + m.fs.micbatt.costing.capital_cost)
            / m.fs.costing.TIC
            / m.fs.feed.properties[0].flow_vol,
            to_units=m.fs.costing.base_currency / (pyunits.m**3 / pyunits.day),
        )
    )
    print(f"Normalized direct capital costs: {DCC_normalized:.2f} $/(m^3/day)")

    ICC_normalized = value(
        pyunits.convert(
            m.fs.costing.total_capital_cost / m.fs.feed.properties[0].flow_vol,
            to_units=m.fs.costing.base_currency / (pyunits.m**3 / pyunits.day),
        )
    )
    print(f"Normalized total capital costs: {ICC_normalized:.2f} $/(m^3/day)")

    total_capital_costs = value(
        pyunits.convert(m.fs.costing.total_capital_cost, to_units=pyunits.USD_2020)
    )
    print(f"Total capital costs: {total_capital_costs:.2f} $")

    print(
        f"Pump capital cost: {value(pyunits.convert(m.fs.pump.costing.capital_cost, to_units=pyunits.USD_2020)):.2f} $"
    )

    print(
        f"Microbial battery capital cost: {value(pyunits.convert(m.fs.micbatt.costing.capital_cost, to_units=pyunits.USD_2020)):.2f} $"
    )

    print("\n------------- Operating costs -------------")
    FMC_normalized = value(
        pyunits.convert(
            m.fs.costing.maintenance_cost / m.fs.costing.total_capital_cost,
            to_units=1 / pyunits.year,
        )
    )
    print(f"Normalized maintenance costs: {FMC_normalized:.2f} 1/year")

    OFOC_normalized = value(
        pyunits.convert(
            m.fs.costing.aggregate_fixed_operating_cost
            / m.fs.costing.total_capital_cost,
            to_units=1 / pyunits.year,
        )
    )
    print(f"Normalized other fixed operating cost: {OFOC_normalized:.2f} 1/year")

    EC_normalized = value(
        pyunits.convert(
            m.fs.costing.aggregate_flow_costs["electricity"]
            * m.fs.costing.utilization_factor
            / m.fs.costing.annual_water_inlet,
            to_units=m.fs.costing.base_currency / pyunits.m**3,
        )
    )
    print(f"Normalized electricity cost: {EC_normalized:.3f} $/m^3 of feed")

    fresh_media_cost_normalized = value(
        pyunits.convert(
            m.fs.costing.aggregate_flow_costs["filtration_media"]
            * m.fs.costing.utilization_factor
            / m.fs.costing.annual_water_inlet,
            to_units=m.fs.costing.base_currency / pyunits.m**3,
        )
    )
    print(
        f"Normalized fresh media cost: {fresh_media_cost_normalized:.3f} $/m^3 of feed"
    )

    spent_media_disposal_cost_normalized = value(
        pyunits.convert(
            m.fs.costing.aggregate_flow_costs["filtration_media_disposal"]
            * m.fs.costing.utilization_factor
            / m.fs.costing.annual_water_inlet,
            to_units=m.fs.costing.base_currency / pyunits.m**3,
        )
    )
    print(
        f"Normalized spent media disposal cost: {spent_media_disposal_cost_normalized:.3f} $/m^3 of feed"
    )

    electricity_intensity = value(
        pyunits.convert(
            m.fs.costing.electricity_intensity, to_units=pyunits.kWh / pyunits.m**3
        )
    )
    print(
        f"Electricity Intensity with respect to influent flowrate: {electricity_intensity:.3f} kWh/m^3"
    )

    total_operating_costs = value(
        pyunits.convert(
            m.fs.costing.total_operating_cost, to_units=pyunits.USD_2020 / pyunits.year
        )
    )
    print(f"Total operating costs: {total_operating_costs:.2f} $/year")

    fixed_operating_costs = value(
        pyunits.convert(
            m.fs.costing.total_fixed_operating_cost,
            to_units=pyunits.USD_2020 / pyunits.year,
        )
    )
    print(f"Fixed operating costs: {fixed_operating_costs:.2f} $/year")

    variable_operating_costs = value(
        pyunits.convert(
            m.fs.costing.total_variable_operating_cost,
            to_units=pyunits.USD_2020 / pyunits.year,
        )
    )
    print(f"Variable operating costs: {variable_operating_costs:.2f} $/year")

    electricity_operating_costs = value(
        pyunits.convert(
            m.fs.costing.aggregate_flow_costs["electricity"]
            * m.fs.costing.utilization_factor,
            to_units=pyunits.USD_2020 / pyunits.year,
        )
    )
    print(f"Electricity operating costs: {electricity_operating_costs:.2f} $/year")

    fresh_media_costs = value(
        pyunits.convert(
            m.fs.costing.aggregate_flow_costs["filtration_media"]
            * m.fs.costing.utilization_factor,
            to_units=pyunits.USD_2020 / pyunits.year,
        )
    )
    print(f"Fresh media operating costs: {fresh_media_costs:.2f} $/year")

    media_disposal_costs = value(
        pyunits.convert(
            m.fs.costing.aggregate_flow_costs["filtration_media_disposal"]
            * m.fs.costing.utilization_factor,
            to_units=pyunits.USD_2020 / pyunits.year,
        )
    )
    print(f"Spent media disposal operating costs: {media_disposal_costs:.2f} $/year")

    print("\n+++++++++++++++++++++++++++++++++++++++++++++++++++++++++")


if __name__ == "__main__":
    m, results = main()



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest
from watertap.property_models.coagulation_prop_pack import CoagulationParameterBlock
from watertap.property_models.NaCl_prop_pack import NaClParameterBlock
from watertap.property_models.seawater_prop_pack import SeawaterParameterBlock
from watertap.unit_models.coag_floc_model import CoagulationFlocculation
from pyomo.environ import (
    ConcreteModel,
    assert_optimal_termination,
    value,
    Param,
    Var,
    units as pyunits,
    Constraint,
)
from idaes.core import FlowsheetBlock
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.model_statistics import degrees_of_freedom
from pyomo.util.check_units import assert_units_consistent
import idaes.core.util.scaling as iscale
from idaes.core.util.testing import initialization_tester
from idaes.core.solvers import get_solver
import re

__author__ = "Austin Ladshaw"

solver = get_solver()

# -----------------------------------------------------------------------------
# Start test class
class TestCoagulation_withChemicals:
    @pytest.fixture(scope="class")
    def coag_obj_w_chems(self):
        model = ConcreteModel()
        model.fs = FlowsheetBlock(dynamic=False)
        model.fs.properties = CoagulationParameterBlock()
        ## NOTE: These values provided are just DUMMY values for the purposes
        #        of testing. They are not meant to be representative of any
        #        particular chemicals or real-world additives.
        chem_dict = {
            "Alum": {
                "parameter_data": {
                    "mw_additive": (200, pyunits.g / pyunits.mol),
                    "moles_salt_per_mole_additive": 3,
                    "mw_salt": (100, pyunits.g / pyunits.mol),
                }
            },
            "Poly": {
                "parameter_data": {
                    "mw_additive": (25, pyunits.g / pyunits.mol),
                    "moles_salt_per_mole_additive": 0,
                    "mw_salt": (23, pyunits.g / pyunits.mol),
                }
            },
        }
        model.fs.unit = CoagulationFlocculation(
            property_package=model.fs.properties, chemical_additives=chem_dict
        )

        return model

    @pytest.mark.unit
    def test_build_model(self, coag_obj_w_chems):
        model = coag_obj_w_chems

        assert len(model.fs.unit.config.chemical_additives) == 2
        assert isinstance(model.fs.unit.slope, Var)
        assert isinstance(model.fs.unit.intercept, Var)
        assert isinstance(model.fs.unit.initial_turbidity_ntu, Var)
        assert isinstance(model.fs.unit.final_turbidity_ntu, Var)
        assert isinstance(model.fs.unit.chemical_doses, Var)
        assert len(model.fs.unit.chemical_doses) == 2
        assert isinstance(model.fs.unit.chemical_mw, Param)
        assert len(model.fs.unit.chemical_mw) == 2
        assert isinstance(model.fs.unit.salt_mw, Param)
        assert len(model.fs.unit.salt_mw) == 2
        assert isinstance(model.fs.unit.salt_from_additive_mole_ratio, Param)
        assert len(model.fs.unit.salt_from_additive_mole_ratio) == 2
        assert isinstance(model.fs.unit.tss_loss_rate, Var)
        assert isinstance(model.fs.unit.eq_tss_loss_rate, Constraint)

        assert isinstance(model.fs.unit.rapid_mixing_retention_time, Var)
        assert isinstance(model.fs.unit.num_rapid_mixing_basins, Var)
        assert isinstance(model.fs.unit.rapid_mixing_vel_grad, Var)

        assert isinstance(model.fs.unit.floc_retention_time, Var)
        assert isinstance(model.fs.unit.single_paddle_length, Var)
        assert isinstance(model.fs.unit.single_paddle_width, Var)
        assert isinstance(model.fs.unit.paddle_rotational_speed, Var)
        assert isinstance(model.fs.unit.paddle_drag_coef, Var)
        assert isinstance(model.fs.unit.vel_fraction, Var)
        assert isinstance(model.fs.unit.num_paddle_wheels, Var)
        assert isinstance(model.fs.unit.num_paddles_per_wheel, Var)

        assert isinstance(model.fs.unit.tds_gain_rate, Var)
        assert isinstance(model.fs.unit.eq_tds_gain_rate, Constraint)

        assert isinstance(model.fs.unit.eq_mass_transfer_term, Constraint)

        assert isinstance(model.fs.unit.eq_rapid_mixing_basin_vol, Constraint)
        assert isinstance(model.fs.unit.rapid_mixing_basin_vol, Var)

        assert isinstance(model.fs.unit.eq_rapid_mixing_power, Constraint)
        assert isinstance(model.fs.unit.rapid_mixing_power, Var)

        assert isinstance(model.fs.unit.eq_floc_basin_vol, Constraint)
        assert isinstance(model.fs.unit.floc_basin_vol, Var)

        assert isinstance(model.fs.unit.eq_floc_wheel_speed, Constraint)
        assert isinstance(model.fs.unit.floc_wheel_speed, Var)

        assert isinstance(model.fs.unit.eq_flocculation_power, Constraint)
        assert isinstance(model.fs.unit.flocculation_power, Var)

        assert isinstance(model.fs.unit.eq_total_power, Constraint)
        assert isinstance(model.fs.unit.total_power, Var)

    @pytest.mark.unit
    def test_stats(self, coag_obj_w_chems):
        model = coag_obj_w_chems

        # Check to make sure we have the correct DOF and
        #   check to make sure the units are correct
        assert_units_consistent(model)
        assert degrees_of_freedom(model) == 22

        # set the operational parameters
        model.fs.unit.fix_tss_turbidity_relation_defaults()
        model.fs.unit.initial_turbidity_ntu.fix()
        model.fs.unit.final_turbidity_ntu.fix(5)
        model.fs.unit.chemical_doses[0, "Alum"].fix(10)
        model.fs.unit.chemical_doses[0, "Poly"].fix(5)

        # set the inlet streams
        assert degrees_of_freedom(model) == 17
        model.fs.unit.inlet.pressure.fix(101325)
        model.fs.unit.inlet.temperature.fix(298.15)
        model.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(1)
        model.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "TDS"].fix(0.01)
        model.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "TSS"].fix(0.01)
        model.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "Sludge"].fix(0.0)

        # set performance vars
        model.fs.unit.rapid_mixing_retention_time[0].fix(60)
        model.fs.unit.num_rapid_mixing_basins.fix(4)
        model.fs.unit.rapid_mixing_vel_grad[0].fix(750)

        model.fs.unit.floc_retention_time[0].fix(1800)
        model.fs.unit.single_paddle_length.fix(4)
        model.fs.unit.single_paddle_width.fix(0.5)
        model.fs.unit.paddle_rotational_speed[0].fix(0.03)

        model.fs.unit.paddle_drag_coef[0].fix(1.5)
        model.fs.unit.vel_fraction.fix(0.7)
        model.fs.unit.num_paddle_wheels.fix(4)
        model.fs.unit.num_paddles_per_wheel.fix(4)

        assert degrees_of_freedom(model) == 0

    @pytest.mark.unit
    def test_scaling(self, coag_obj_w_chems):
        model = coag_obj_w_chems

        # Set some scaling factors and look for 'bad' scaling
        model.fs.properties.set_default_scaling(
            "flow_mass_phase_comp", 1, index=("Liq", "H2O")
        )
        model.fs.properties.set_default_scaling(
            "flow_mass_phase_comp", 1e2, index=("Liq", "TSS")
        )
        model.fs.properties.set_default_scaling(
            "flow_mass_phase_comp", 1e2, index=("Liq", "TDS")
        )
        model.fs.properties.set_default_scaling(
            "flow_mass_phase_comp", 1e2, index=("Liq", "Sludge")
        )

        # set scaling factors for performance
        iscale.set_scaling_factor(model.fs.unit.rapid_mixing_retention_time, 1e-1)
        iscale.set_scaling_factor(model.fs.unit.num_rapid_mixing_basins, 1)
        iscale.set_scaling_factor(model.fs.unit.rapid_mixing_vel_grad, 1e-2)

        iscale.set_scaling_factor(model.fs.unit.floc_retention_time, 1e-3)
        iscale.set_scaling_factor(model.fs.unit.single_paddle_length, 1)
        iscale.set_scaling_factor(model.fs.unit.single_paddle_width, 1)
        iscale.set_scaling_factor(model.fs.unit.paddle_rotational_speed, 10)

        iscale.set_scaling_factor(model.fs.unit.paddle_drag_coef, 1)
        iscale.set_scaling_factor(model.fs.unit.vel_fraction, 1)
        iscale.set_scaling_factor(model.fs.unit.num_paddle_wheels, 1)
        iscale.set_scaling_factor(model.fs.unit.num_paddles_per_wheel, 1)

        iscale.calculate_scaling_factors(model.fs)

        # check that all variables have scaling factors
        unscaled_var_list = list(iscale.unscaled_variables_generator(model))
        assert len(unscaled_var_list) == 0

        # check if any variables are badly scaled
        badly_scaled_var_values = {
            var.name: val
            for (var, val) in iscale.badly_scaled_var_generator(
                model, large=1e2, small=1e-2
            )
        }
        assert not badly_scaled_var_values

    @pytest.mark.component
    def test_initialization(self, coag_obj_w_chems):
        model = coag_obj_w_chems
        initialization_tester(model)

        # check to make sure DOF does not change
        assert degrees_of_freedom(model) == 0

    @pytest.mark.component
    def test_solve(self, coag_obj_w_chems):
        model = coag_obj_w_chems

        # first, check to make sure that after initialized, the scaling is still good
        badly_scaled_var_values = {
            var.name: val
            for (var, val) in iscale.badly_scaled_var_generator(
                model, large=1e2, small=1e-2
            )
        }
        assert not badly_scaled_var_values

        # run solver and check for optimal solution
        results = solver.solve(model)
        assert_optimal_termination(results)

    @pytest.mark.component
    def test_solution(self, coag_obj_w_chems):
        model = coag_obj_w_chems

        assert value(
            model.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "H2O"]
        ) == pytest.approx(1, rel=1e-4)
        assert value(
            model.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "Sludge"]
        ) == pytest.approx(0.00999, rel=1e-4)
        assert value(
            model.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "TDS"]
        ) == pytest.approx(0.010015, rel=1e-4)
        assert value(
            model.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "TSS"]
        ) == pytest.approx(9.36352e-06, rel=1e-4)

    @pytest.mark.component
    def test_performance_contents(self, coag_obj_w_chems):
        model = coag_obj_w_chems

        dict = model.fs.unit._get_performance_contents()

        assert "vars" in dict
        assert "Total Power Usage  (kW)" in dict["vars"]
        assert "Rapid Mixing Power (kW)" in dict["vars"]
        assert "Flocc Mixing Power (kW)" in dict["vars"]

        assert value(dict["vars"]["Total Power Usage  (kW)"]) == pytest.approx(
            0.56798, rel=1e-4
        )
        assert value(dict["vars"]["Rapid Mixing Power (kW)"]) == pytest.approx(
            0.12115, rel=1e-4
        )
        assert value(dict["vars"]["Flocc Mixing Power (kW)"]) == pytest.approx(
            0.44684, rel=1e-4
        )


# -----------------------------------------------------------------------------
# Start test class without chemicals added
class TestCoagulation_withNoChemicals:
    @pytest.fixture(scope="class")
    def coag_obj_wo_chems(self):
        model = ConcreteModel()
        model.fs = FlowsheetBlock(dynamic=False)
        model.fs.properties = CoagulationParameterBlock()
        model.fs.unit = CoagulationFlocculation(property_package=model.fs.properties)

        return model

    @pytest.mark.unit
    def test_build_model(self, coag_obj_wo_chems):
        model = coag_obj_wo_chems

        assert len(model.fs.unit.config.chemical_additives) == 0
        assert isinstance(model.fs.unit.slope, Var)
        assert isinstance(model.fs.unit.intercept, Var)
        assert isinstance(model.fs.unit.initial_turbidity_ntu, Var)
        assert isinstance(model.fs.unit.final_turbidity_ntu, Var)
        assert isinstance(model.fs.unit.chemical_doses, Var)
        assert len(model.fs.unit.chemical_doses) == 0
        assert isinstance(model.fs.unit.chemical_mw, Param)
        assert len(model.fs.unit.chemical_mw) == 0
        assert isinstance(model.fs.unit.salt_mw, Param)
        assert len(model.fs.unit.salt_mw) == 0
        assert isinstance(model.fs.unit.salt_from_additive_mole_ratio, Param)
        assert len(model.fs.unit.salt_from_additive_mole_ratio) == 0
        assert isinstance(model.fs.unit.tss_loss_rate, Var)
        assert isinstance(model.fs.unit.eq_tss_loss_rate, Constraint)

        assert not hasattr(model.fs.unit, "tds_gain_rate")
        assert not hasattr(model.fs.unit, "eq_tds_gain_rate")

        assert isinstance(model.fs.unit.eq_mass_transfer_term, Constraint)

    @pytest.mark.unit
    def test_stats(self, coag_obj_wo_chems):
        model = coag_obj_wo_chems

        # Check to make sure we have the correct DOF and
        #   check to make sure the units are correct
        assert_units_consistent(model)
        assert degrees_of_freedom(model) == 20

        # set the operational parameters
        model.fs.unit.fix_tss_turbidity_relation_defaults()
        model.fs.unit.initial_turbidity_ntu.fix(5000)
        model.fs.unit.final_turbidity_ntu.fix(100)

        # set the inlet streams
        assert degrees_of_freedom(model) == 17

        tss_in = value(model.fs.unit.compute_inlet_tss_mass_flow(0))
        assert tss_in == pytest.approx(0.0093, rel=1e-4)
        model.fs.unit.inlet.pressure.fix(101325)
        model.fs.unit.inlet.temperature.fix(298.15)
        model.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(1)
        model.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "TDS"].fix(0.01)
        model.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "TSS"].fix(tss_in)
        model.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "Sludge"].fix(0.0)

        # set performance vars
        model.fs.unit.rapid_mixing_retention_time[0].fix(60)
        model.fs.unit.num_rapid_mixing_basins.fix(4)
        model.fs.unit.rapid_mixing_vel_grad[0].fix(750)

        model.fs.unit.floc_retention_time[0].fix(1800)
        model.fs.unit.single_paddle_length.fix(4)
        model.fs.unit.single_paddle_width.fix(0.5)
        model.fs.unit.paddle_rotational_speed[0].fix(0.03)

        model.fs.unit.paddle_drag_coef[0].fix(1.5)
        model.fs.unit.vel_fraction.fix(0.7)
        model.fs.unit.num_paddle_wheels.fix(4)
        model.fs.unit.num_paddles_per_wheel.fix(4)

        assert degrees_of_freedom(model) == 0

    @pytest.mark.unit
    def test_scaling(self, coag_obj_wo_chems):
        model = coag_obj_wo_chems

        # Set some scaling factors and look for 'bad' scaling
        model.fs.properties.set_default_scaling(
            "flow_mass_phase_comp", 1, index=("Liq", "H2O")
        )
        model.fs.properties.set_default_scaling(
            "flow_mass_phase_comp", 1e2, index=("Liq", "TSS")
        )
        model.fs.properties.set_default_scaling(
            "flow_mass_phase_comp", 1e2, index=("Liq", "TDS")
        )
        model.fs.properties.set_default_scaling(
            "flow_mass_phase_comp", 1e2, index=("Liq", "Sludge")
        )

        # Here we are skipping setting scaling factors for performance to test the
        #   effectiveness of the defaults AND get better test coverage

        iscale.calculate_scaling_factors(model.fs)

        # check that all variables have scaling factors
        unscaled_var_list = list(iscale.unscaled_variables_generator(model))
        assert len(unscaled_var_list) == 0

        # check if any variables are badly scaled
        badly_scaled_var_values = {
            var.name: val
            for (var, val) in iscale.badly_scaled_var_generator(
                model, large=1e2, small=1e-2
            )
        }
        print(iscale.get_scaling_factor(model.fs.unit.tss_loss_rate))
        assert not badly_scaled_var_values

    @pytest.mark.component
    def test_initialization(self, coag_obj_wo_chems):
        model = coag_obj_wo_chems
        initialization_tester(model)

        # check to make sure DOF does not change
        assert degrees_of_freedom(model) == 0

    @pytest.mark.component
    def test_solve(self, coag_obj_wo_chems):
        model = coag_obj_wo_chems

        # first, check to make sure that after initialized, the scaling is still good
        badly_scaled_var_values = {
            var.name: val
            for (var, val) in iscale.badly_scaled_var_generator(
                model, large=1e2, small=1e-2
            )
        }
        assert not badly_scaled_var_values

        # run solver and check for optimal solution
        results = solver.solve(model)
        assert_optimal_termination(results)

    @pytest.mark.component
    def test_solution(self, coag_obj_wo_chems):
        model = coag_obj_wo_chems

        assert value(
            model.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "H2O"]
        ) == pytest.approx(1, rel=1e-4)
        assert value(
            model.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "Sludge"]
        ) == pytest.approx(0.0091127, rel=1e-4)
        assert value(
            model.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "TDS"]
        ) == pytest.approx(0.01, rel=1e-4)
        assert value(
            model.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "TSS"]
        ) == pytest.approx(0.00018725, rel=1e-4)


# -----------------------------------------------------------------------------
# Start test class with bad config
class TestCoagulation_withBadConfig:
    @pytest.fixture(scope="class")
    def coag_obj_bad_config(self):
        model = ConcreteModel()
        model.fs = FlowsheetBlock(dynamic=False)
        model.fs.properties = CoagulationParameterBlock()

        return model

    @pytest.mark.unit
    def test_build_model_catch_errors(self, coag_obj_bad_config):
        model = coag_obj_bad_config

        bad_dict1 = {
            "Alum": {
                "foo_bar": {
                    "mw_additive": (200, pyunits.g / pyunits.mol),
                    "moles_salt_per_mole_additive": 3,
                    "mw_salt": (100, pyunits.g / pyunits.mol),
                }
            }
        }
        with pytest.raises(
            ConfigurationError, match="Did not provide a 'parameter_data' for chemical"
        ):
            model.fs.unit = CoagulationFlocculation(
                property_package=model.fs.properties, chemical_additives=bad_dict1
            )

        bad_dict2 = {
            "Alum": {
                "parameter_data": {
                    "foo_bar": (200, pyunits.g / pyunits.mol),
                    "moles_salt_per_mole_additive": 3,
                    "mw_salt": (100, pyunits.g / pyunits.mol),
                }
            }
        }
        with pytest.raises(
            ConfigurationError, match="Did not provide a 'mw_additive' for chemical"
        ):
            model.fs.unit = CoagulationFlocculation(
                property_package=model.fs.properties, chemical_additives=bad_dict2
            )

        bad_dict3 = {
            "Alum": {
                "parameter_data": {
                    "mw_additive": (200, pyunits.g / pyunits.mol),
                    "moles_salt_per_mole_additive": "foo-bar",
                    "mw_salt": (100, pyunits.g / pyunits.mol),
                }
            }
        }
        with pytest.raises(
            ConfigurationError,
            match="Did not provide a number for 'moles_salt_per_mole_additive'",
        ):
            model.fs.unit = CoagulationFlocculation(
                property_package=model.fs.properties, chemical_additives=bad_dict3
            )

        bad_dict4 = {
            "Alum": {
                "parameter_data": {
                    "mw_additive": (200, pyunits.g / pyunits.mol),
                    "foo-bar": 3,
                    "mw_salt": (100, pyunits.g / pyunits.mol),
                }
            }
        }
        with pytest.raises(
            ConfigurationError,
            match="Did not provide a 'moles_salt_per_mole_additive' for chemical",
        ):
            model.fs.unit = CoagulationFlocculation(
                property_package=model.fs.properties, chemical_additives=bad_dict4
            )

        bad_dict5 = {
            "Alum": {
                "parameter_data": {
                    "mw_additive": (200, pyunits.g / pyunits.mol),
                    "moles_salt_per_mole_additive": 3,
                    "foo-bar": (100, pyunits.g / pyunits.mol),
                }
            }
        }
        with pytest.raises(
            ConfigurationError, match="Did not provide a 'mw_salt' for chemical"
        ):
            model.fs.unit = CoagulationFlocculation(
                property_package=model.fs.properties, chemical_additives=bad_dict5
            )

        bad_dict6 = {
            "Alum": {
                "parameter_data": {
                    "mw_additive": "not a tuple",
                    "moles_salt_per_mole_additive": 3,
                    "mw_salt": (100, pyunits.g / pyunits.mol),
                }
            }
        }
        with pytest.raises(
            ConfigurationError, match="Did not provide a tuple for 'mw_additive'"
        ):
            model.fs.unit = CoagulationFlocculation(
                property_package=model.fs.properties, chemical_additives=bad_dict6
            )


# -----------------------------------------------------------------------------
# Start test class with bad config
class TestCoagulation_withBadProperties:
    @pytest.fixture(scope="class")
    def coag_obj_bad_properties(self):
        model = ConcreteModel()
        model.fs = FlowsheetBlock(dynamic=False)

        return model

    @pytest.mark.unit
    def test_build_model_catch_prop_errors(self, coag_obj_bad_properties):
        model = coag_obj_bad_properties

        error_msg = (
            "Coagulation-Flocculation model MUST contain ('Liq','TDS') "
            "as a component, but the property package has only specified "
            "the following components [('Liq', 'H2O'), ('Liq', 'NaCl')]"
        )
        with pytest.raises(ConfigurationError, match=re.escape(error_msg)):
            model.fs.properties = NaClParameterBlock()
            model.fs.unit = CoagulationFlocculation(
                property_package=model.fs.properties
            )

        error_msg = (
            "Coagulation-Flocculation model MUST contain ('Liq','Sludge') "
            "as a component, but the property package has only specified "
            "the following components [('Liq', 'H2O'), ('Liq', 'TDS')]"
        )
        with pytest.raises(ConfigurationError, match=re.escape(error_msg)):
            model.fs.properties = SeawaterParameterBlock()
            model.fs.unit = CoagulationFlocculation(
                property_package=model.fs.properties
            )

        # NOTE: package must also contain ('Liq','TSS') as a component



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for zero-order sedimentation model
"""
import pytest


from pyomo.environ import (
    Block,
    ConcreteModel,
    Constraint,
    value,
    Var,
    assert_optimal_termination,
)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.core import UnitModelCostingBlock

from watertap.unit_models.zero_order import SedimentationZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.costing.zero_order_costing import ZeroOrderCosting

solver = get_solver()


class TestSedimentationZO_w_default_removal:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(solute_list=["tss", "foo"])

        m.fs.unit = SedimentationZO(property_package=m.fs.params, database=m.db)

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(3)
        m.fs.unit.inlet.flow_mass_comp[0, "foo"].fix(1)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db

        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)

        assert isinstance(model.fs.unit.electricity_consumption, Constraint)

        assert isinstance(model.fs.unit.settling_velocity, Var)
        assert isinstance(model.fs.unit.basin_surface_area, Var)
        assert isinstance(model.fs.unit.basin_surface_area_constraint, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("sedimentation")

        model.fs.unit.load_parameters_from_database(use_default_removal=True)

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert (
            model.fs.unit.recovery_frac_mass_H2O[0].value
            == data["recovery_frac_mass_H2O"]["value"]
        )

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            assert v.fixed
            if j == "foo":
                assert v.value == data["default_removal_frac_mass_comp"]["value"]
            else:
                assert v.value == data["removal_frac_mass_comp"][j]["value"]

        assert model.fs.unit.energy_electric_flow_vol_inlet.fixed
        assert (
            model.fs.unit.energy_electric_flow_vol_inlet.value
            == data["energy_electric_flow_vol_inlet"]["value"]
        )

        assert model.fs.unit.settling_velocity[0].fixed
        assert (
            model.fs.unit.settling_velocity[0].value
            == data["settling_velocity"]["value"]
        )

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert pytest.approx(1.4e-2, rel=1e-5) == value(
            model.fs.unit.properties_in[0].flow_vol
        )
        assert pytest.approx(214.2857, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["tss"]
        )
        assert pytest.approx(71.4286, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["foo"]
        )
        assert pytest.approx(1.10265e-2, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].flow_vol
        )
        assert pytest.approx(2.4960, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["tss"]
        )
        assert pytest.approx(90.690, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["foo"]
        )
        assert pytest.approx(2.9735e-3, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].flow_vol
        )
        assert pytest.approx(999.66, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["tss"]
        )
        assert pytest.approx(0, abs=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["foo"]
        )
        assert pytest.approx(0, abs=1e-5) == value(model.fs.unit.electricity[0])
        assert pytest.approx(30.139, rel=1e-5) == value(
            model.fs.unit.basin_surface_area[0]
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(
                value(
                    model.fs.unit.inlet.flow_mass_comp[0, j]
                    - model.fs.unit.treated.flow_mass_comp[0, j]
                    - model.fs.unit.byproduct.flow_mass_comp[0, j]
                )
            )

    @pytest.mark.component
    def test_report(self, model):

        model.fs.unit.report()


class TestSedimentationZO_phosphorus_capture_tss:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(solute_list=["tss"])

        m.fs.unit = SedimentationZO(
            property_package=m.fs.params,
            database=m.db,
            process_subtype="phosphorus_capture",
        )

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(3)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db

        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)

        assert isinstance(model.fs.unit.electricity_consumption, Constraint)

        assert isinstance(model.fs.unit.settling_velocity, Var)
        assert isinstance(model.fs.unit.basin_surface_area, Var)
        assert isinstance(model.fs.unit.basin_surface_area_constraint, Constraint)
        assert isinstance(model.fs.unit.final_phosphate_mass, Var)
        assert isinstance(model.fs.unit.phosphate_mass_flow_constraint, Constraint)
        assert isinstance(model.fs.unit.phosphorus_solids_ratio, Var)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters(
            "sedimentation", subtype="phosphorus_capture"
        )

        model.fs.unit.load_parameters_from_database()

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert (
            model.fs.unit.recovery_frac_mass_H2O[0].value
            == data["recovery_frac_mass_H2O"]["value"]
        )

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            assert v.fixed
            assert v.value == data["removal_frac_mass_comp"][j]["value"]

        assert model.fs.unit.energy_electric_flow_vol_inlet.fixed
        assert (
            model.fs.unit.energy_electric_flow_vol_inlet.value
            == data["energy_electric_flow_vol_inlet"]["value"]
        )

        assert model.fs.unit.settling_velocity[0].fixed
        assert (
            model.fs.unit.settling_velocity[0].value
            == data["settling_velocity"]["value"]
        )

        assert model.fs.unit.phosphorus_solids_ratio[0].fixed
        assert (
            model.fs.unit.phosphorus_solids_ratio[0].value
            == data["phosphorus_solids_ratio"]["value"]
        )

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert pytest.approx(0.009864, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].flow_vol
        )
        assert pytest.approx(0.3041363, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["tss"]
        )
        assert pytest.approx(0.0031360, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].flow_vol
        )
        assert pytest.approx(955.68, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["tss"]
        )
        assert pytest.approx(0, abs=1e-5) == value(model.fs.unit.electricity[0])
        assert pytest.approx(27.986, rel=1e-5) == value(
            model.fs.unit.basin_surface_area[0]
        )
        assert pytest.approx(0.44955, rel=1e-5) == value(
            model.fs.unit.final_phosphate_mass[0]
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(
                value(
                    model.fs.unit.inlet.flow_mass_comp[0, j]
                    - model.fs.unit.treated.flow_mass_comp[0, j]
                    - model.fs.unit.byproduct.flow_mass_comp[0, j]
                )
            )

    @pytest.mark.component
    def test_report(self, model):

        model.fs.unit.report()


class TestSedimentationZO_phosphorus_capture_phosphates:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(solute_list=["phosphates"])

        m.fs.unit = SedimentationZO(
            property_package=m.fs.params,
            database=m.db,
            process_subtype="phosphorus_capture",
        )

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10)
        m.fs.unit.inlet.flow_mass_comp[0, "phosphates"].fix(3)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db

        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)

        assert isinstance(model.fs.unit.electricity_consumption, Constraint)

        assert isinstance(model.fs.unit.settling_velocity, Var)
        assert isinstance(model.fs.unit.basin_surface_area, Var)
        assert isinstance(model.fs.unit.basin_surface_area_constraint, Constraint)
        assert isinstance(model.fs.unit.final_solids_mass, Var)
        assert isinstance(model.fs.unit.solids_mass_flow_constraint, Constraint)
        assert isinstance(model.fs.unit.phosphorus_solids_ratio, Var)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters(
            "sedimentation", subtype="phosphorus_capture"
        )

        model.fs.unit.load_parameters_from_database()

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert (
            model.fs.unit.recovery_frac_mass_H2O[0].value
            == data["recovery_frac_mass_H2O"]["value"]
        )

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            assert v.fixed
            assert v.value == data["removal_frac_mass_comp"][j]["value"]

        assert model.fs.unit.energy_electric_flow_vol_inlet.fixed
        assert (
            model.fs.unit.energy_electric_flow_vol_inlet.value
            == data["energy_electric_flow_vol_inlet"]["value"]
        )

        assert model.fs.unit.settling_velocity[0].fixed
        assert (
            model.fs.unit.settling_velocity[0].value
            == data["settling_velocity"]["value"]
        )

        assert model.fs.unit.phosphorus_solids_ratio[0].fixed
        assert (
            model.fs.unit.phosphorus_solids_ratio[0].value
            == data["phosphorus_solids_ratio"]["value"]
        )

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert pytest.approx(0.009864, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].flow_vol
        )
        assert pytest.approx(0.3041363, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["phosphates"]
        )
        assert pytest.approx(0.0031360, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].flow_vol
        )
        assert pytest.approx(955.68, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["phosphates"]
        )
        assert pytest.approx(0, abs=1e-5) == value(model.fs.unit.electricity[0])
        assert pytest.approx(27.986, rel=1e-5) == value(
            model.fs.unit.basin_surface_area[0]
        )
        assert pytest.approx(19.980, rel=1e-5) == value(
            model.fs.unit.final_solids_mass[0]
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(
                value(
                    model.fs.unit.inlet.flow_mass_comp[0, j]
                    - model.fs.unit.treated.flow_mass_comp[0, j]
                    - model.fs.unit.byproduct.flow_mass_comp[0, j]
                )
            )

    @pytest.mark.component
    def test_report(self, model):

        model.fs.unit.report()


db = Database()
params = db._get_technology("sedimentation")


class TestSedimentationZOsubtype:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(solute_list=["tds", "tss"])

        m.fs.unit = SedimentationZO(property_package=m.fs.params, database=db)

        return m

    @pytest.mark.parametrize("subtype", [k for k in params.keys()])
    @pytest.mark.component
    def test_load_parameters(self, model, subtype):
        model.fs.unit.config.process_subtype = subtype
        data = db.get_unit_operation_parameters("sedimentation", subtype=subtype)

        model.fs.unit.load_parameters_from_database(use_default_removal=True)

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            if j not in data["removal_frac_mass_comp"].keys():
                assert v.value == data["default_removal_frac_mass_comp"]["value"]
            else:
                assert v.value == data["removal_frac_mass_comp"][j]["value"]


@pytest.mark.parametrize(
    "subtype", [k for k in params.keys() if k != "phosphorus_capture"]
)
@pytest.mark.component
def test_costing(subtype):
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = WaterParameterBlock(solute_list=["sulfur", "toc", "tss"])

    m.fs.costing = ZeroOrderCosting()

    m.fs.unit1 = SedimentationZO(property_package=m.fs.params, database=m.db)

    m.fs.unit1.inlet.flow_mass_comp[0, "H2O"].fix(10000)
    m.fs.unit1.inlet.flow_mass_comp[0, "sulfur"].fix(1)
    m.fs.unit1.inlet.flow_mass_comp[0, "toc"].fix(2)
    m.fs.unit1.inlet.flow_mass_comp[0, "tss"].fix(3)
    m.fs.unit1.load_parameters_from_database(use_default_removal=True)
    assert degrees_of_freedom(m.fs.unit1) == 0

    m.fs.unit1.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    assert isinstance(m.fs.costing.sedimentation, Block)
    assert isinstance(m.fs.costing.sedimentation.capital_a_parameter, Var)
    assert isinstance(m.fs.costing.sedimentation.capital_b_parameter, Var)

    assert isinstance(m.fs.unit1.costing.capital_cost, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost_constraint, Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit1) == 0

    assert m.fs.unit1.electricity[0] in m.fs.costing._registered_flows["electricity"]


@pytest.mark.component
def test_costing_phosphorus_capture():
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = WaterParameterBlock(solute_list=["sulfur", "toc", "tss"])

    m.fs.costing = ZeroOrderCosting()

    m.fs.unit1 = SedimentationZO(
        property_package=m.fs.params,
        database=m.db,
        process_subtype="phosphorus_capture",
    )

    m.fs.unit1.inlet.flow_mass_comp[0, "H2O"].fix(10000)
    m.fs.unit1.inlet.flow_mass_comp[0, "sulfur"].fix(1)
    m.fs.unit1.inlet.flow_mass_comp[0, "toc"].fix(2)
    m.fs.unit1.inlet.flow_mass_comp[0, "tss"].fix(3)
    m.fs.unit1.load_parameters_from_database(use_default_removal=True)
    assert degrees_of_freedom(m.fs.unit1) == 0

    m.fs.unit1.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    assert isinstance(m.fs.costing.sedimentation, Block)
    assert isinstance(m.fs.costing.sedimentation.unit_capex, Var)
    assert isinstance(m.fs.costing.sedimentation.unit_opex, Var)

    assert isinstance(m.fs.unit1.costing.capital_cost, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost_constraint, Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit1) == 0

    assert m.fs.unit1.electricity[0] in m.fs.costing._registered_flows["electricity"]


@pytest.mark.unit
def test_phosphorus_capture_no_tss_or_phosphate_in_solute_list_error():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.params = WaterParameterBlock(solute_list=["foo"])

    with pytest.raises(
        KeyError,
        match="One of the following should be specified in the solute_list: "
        "tss or phosphates.",
    ):
        m.fs.unit = SedimentationZO(
            property_package=m.fs.params,
            database=db,
            process_subtype="phosphorus_capture",
        )


@pytest.mark.unit
def test_phosphorus_capture_phosphate_tss_in_solute_list_error():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.params = WaterParameterBlock(solute_list=["tss", "phosphates"])

    with pytest.raises(
        KeyError,
        match="tss and phosphates cannot both be defined in the "
        "solute_list. Please choose one.",
    ):
        m.fs.unit = SedimentationZO(
            property_package=m.fs.params,
            database=db,
            process_subtype="phosphorus_capture",
        )
